################################################################
# config.py
#
# configuration class for the aerocom python tools 
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171106 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

#Copyright (C) 2017 met.no
#Contact information:
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA

"""
Provides access to pyaerocom specific configuration values
"""

import numpy as np
import os
import getpass
from collections import OrderedDict as od
import pyaerocom.obs_io as obs_io
from pyaerocom.grid_io import GridIO
from pyaerocom._lowlevel_helpers import (list_to_shortstr, 
                                         chk_make_subdir,
                                         check_fun_timeout_multiproc)
from pyaerocom.variable import VarCollection
try:
    from ConfigParser import ConfigParser
except: 
    from configparser import ConfigParser
    

class Config(object):
    """Class containing relevant paths for read and write routines
    
    TODO: write docstring
    """
    
    # NAMES
    # default names of the different obs networks
    # might get overwritten from paths.ini see func read_config
    
    #: Aeronet Sun V2 access names
    AERONET_SUN_V2L15_AOD_DAILY_NAME = 'AeronetSunV2Lev1.5.daily'
    AERONET_SUN_V2L15_AOD_ALL_POINTS_NAME = 'AeronetSun_2.0_NRT'
    AERONET_SUN_V2L2_AOD_DAILY_NAME = 'AeronetSunV2Lev2.daily'
    AERONET_SUN_V2L2_AOD_ALL_POINTS_NAME = 'AeronetSunV2Lev2.AP'
    
    #: Aeronet SDA V2 access names
    AERONET_SUN_V2L2_SDA_DAILY_NAME = 'AeronetSDAV2Lev2.daily'
    AERONET_SUN_V2L2_SDA_ALL_POINTS_NAME = 'AeronetSDAV2Lev2.AP'
    
    # Aeronet V2 inversion products
    AERONET_INV_V2L15_DAILY_NAME = 'AeronetInvV2Lev1.5.daily'
    AERONET_INV_V2L15_ALL_POINTS_NAME = 'AeronetInvV2Lev1.5.AP'
    AERONET_INV_V2L2_DAILY_NAME = 'AeronetInvV2Lev2.daily'
    AERONET_INV_V2L2_ALL_POINTS_NAME = 'AeronetInvV2Lev2.AP'
    
    #: Aeronet Sun V3 access names
    AERONET_SUN_V3L15_AOD_DAILY_NAME = 'AeronetSunV3Lev1.5.daily'
    AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME = 'AeronetSunV3Lev1.5.AP'
    AERONET_SUN_V3L2_AOD_DAILY_NAME = 'AeronetSunV3Lev2.daily'
    AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME = 'AeronetSunV3Lev2.AP'
    
    #: Aeronet SDA V3 access names
    AERONET_SUN_V3L15_SDA_DAILY_NAME = 'AeronetSDAV3Lev1.5.daily'
    AERONET_SUN_V3L15_SDA_ALL_POINTS_NAME = 'AeronetSDAV3Lev1.5.AP'
    AERONET_SUN_V3L2_SDA_DAILY_NAME = 'AeronetSDAV3Lev2.daily'
    AERONET_SUN_V3L2_SDA_ALL_POINTS_NAME = 'AeronetSDAV3Lev2.AP'

    #: Aeronet V3 inversions
    AERONET_INV_V3L15_DAILY_NAME = 'AeronetInvV3Lev1.5.daily'
    AERONET_INV_V3L2_DAILY_NAME = 'AeronetInvV3Lev2.daily'
    
    #: EBAS name
    EBAS_MULTICOLUMN_NAME = 'EBASMC'
    
    #: EEA nmea
    EEA_NAME = 'EEAAQeRep'

    #: Earlinet access name;
    EARLINET_NAME = 'EARLINET'

    #: GAW TAD subset aas et al paper
    GAWTADSUBSETAASETAL_NAME = 'GAWTADsubsetAasEtAl'

    #: DMS
    DMS_AMS_CVO_NAME = 'DMS_AMS_CVO'

    #: Lowest possible year in data
    MIN_YEAR = 0
    #: Highest possible year in data
    MAX_YEAR = 20000

    #: standard names for coordinates
    STANDARD_COORD_NAMES = ['latitude',
                            'longitude',
                            'altitude']
    #: Information specifying default vertical grid for post processing of
    #: profile data. The values are in units of m.
    DEFAULT_VERT_GRID_DEF = od(lower = 0,
                               upper = 15000,
                               step  = 250)
    #: maximum allowed RH to be considered dry
    RH_MAX_PERCENT_DRY = 40
    
    #: If True, then whenever applicable the time resampling constraints
    #: definted below (OBS_MIN_NUM_RESMAMPLE) are applied to observations when 
    #: resampling in StationData and thus colocation routines. Requires that 
    #: original obs_data is available in a certain regular resolution (or at
    #: least has ts_type assigned to it)
    OBS_APPLY_TIME_RESAMPLE_CONSTRAINTS = True
    
    #: Time resample strategies for certain cominations, first level refers
    #: to TO, second to FROM and values are minimum number of observations
    OBS_MIN_NUM_RESAMPLE = dict(yearly      =   dict(monthly    = 3),
                                monthly     =   dict(daily      = 7),
                                daily       =   dict(hourly     = 6),
                                hourly      =   dict(minutely   = 15))
    
    #: This boolean can be used to enable / disable the former (i.e. use
    #: available wavelengths of variable in a certain range around variable
    #: wavelength).
    OBS_ALLOW_ALT_WAVELENGTHS = obs_io.OBS_ALLOW_ALT_WAVELENGTHS
    
    #: Wavelength tolerance for observations imports
    OBS_WAVELENGTH_TOL_NM = obs_io.OBS_WAVELENGTH_TOL_NM
    
    #: not used at the moment
    GCOSPERCENTCRIT =   np.float(0.1)
    GCOSABSCRIT     =   np.float(0.04)
    
    #names of the different obs networks
    OBSNET_NONE = 'NONE'
    NOMODELNAME = 'OBSERVATIONS-ONLY'

    # names for the satellite data sets
    SENTINEL5P_NAME = 'Sentinel5P'
    AEOLUS_NAME = 'AeolusL2A'
    
    RM_CACHE_OUTDATED = True

    #: Name of the file containing the revision string of an obs data network
    REVISION_FILE = 'Revision.txt'

    
    BASEDIR_PPI = os.path.join('/lustre', 'storeA', 'project', 'aerocom') 
    BASEDIR_USER_SERVER = os.path.join('/metno', 'aerocom-users-database')
    
    #: timeout to check if one of the supported server locations can be 
    #: accessed
    SERVER_CHECK_TIMEOUT = 5 #0.1 #s
    
    from pyaerocom import __dir__
    _config_ini = os.path.join(__dir__, 'data', 'paths.ini')
    _config_ini_user_server = os.path.join(__dir__, 'data', 'paths_user_server.ini')
    _config_ini_testdata = os.path.join(__dir__, 'data', 'paths_testdata.ini')
    
    _config_files = {'metno'                  : _config_ini,
                     'aerocom-users-database' : _config_ini_user_server,
                     'pyaerocom-testdata'     : _config_ini_testdata}
    
    _var_info_file = os.path.join(__dir__, 'data', 'variables.ini')
    _coords_info_file = os.path.join(__dir__, 'data', 'coords.ini')
    _outhomename = 'MyPyaerocom'
    
    DONOTCACHEFILE = None
    def __init__(self, model_base_dir=None, obs_base_dir=None, 
                 output_dir=None, config_file=None, 
                 cache_dir=None, colocateddata_dir=None,
                 write_fileio_err_log=True, 
                 activate_caching=True):
        
        # Loggers
        from pyaerocom import print_log, logger
        self.print_log = print_log
        self.logger = logger
        
        # Directories
        self._modelbasedir = model_base_dir
        self._obsbasedir = obs_base_dir
        self._cachedir = cache_dir
        self._outputdir = output_dir
        self._testdatadir = os.path.join(self.HOMEDIR, 'pyaerocom-testdata')
        self._colocateddatadir = colocateddata_dir
        
        # Options
        self._caching_active = activate_caching
        
        #: Settings for reading and writing of gridded data
        self.GRID_IO = GridIO()
        print_log.info('Initating pyaerocom configuration')
        
        
        if not isinstance(config_file, str) or not os.path.exists(config_file):
            from time import time
            print_log.info('Checking database access...')
            t0 = time()
            config_file = self._infer_config_file()
            print_log.info('Expired time: {:.3f} s'.format(time() - t0))
        
        
        self._var_param = None
        self._coords = None
        
        # Attributes that are used to store search directories
        self.OBSCONFIG = od()
        self.SUPPLDIRS = od()
        self.MODELDIRS = []
        
        self.WRITE_FILEIO_ERR_LOG = write_fileio_err_log
        
        
        self._ebas_flag_info = None
        
        if config_file is not None:
            
            keep_basedirs = False
            if self.dir_exists(model_base_dir) and self.dir_exists(obs_base_dir):
                keep_basedirs=True
            try:
                self.read_config(config_file, keep_basedirs)
                
            except Exception as e:
                from traceback import format_exc
                self.init_outputdirs()
                self.print_log.warning(format_exc())
                self.print_log.warning("Failed to init config. Error: %s" %repr(e))
        else:
            self.init_outputdirs()
    
    def _check_access(self, loc):
        """Uses multiprocessing approach to check if location can be accessed
        
        Wrapper for :func:`check_fun_timeout_multiproc`. Uses method 
        :func:`os.listdir` to validate accessibility at input location
        (without the timeout and for mounted remote locations, this could 
        otherwise take ages)
        
        Parameters
        ----------
        loc : str
            path that is supposed to be checked
        
        Returns
        -------
        bool
            True, if location is accessible, else False
        """
# =============================================================================
#         try:
#             os.listdir(loc)
#             return True
#         except:
#             return False
# =============================================================================
        self.print_log.info('Checking access to: {}'.format(loc))
        return check_fun_timeout_multiproc(os.listdir, fun_args=(loc, ),
                                           timeout_secs=self.SERVER_CHECK_TIMEOUT)
    
    @property
    def has_access_lustre(self):
        """Boolean specifying whether the lustre database can be accessed"""
        ok = False
        p = os.path.join(self.ROOTDIR, 'lustre/')
        if os.path.exists(p):
            ok = self._check_access(os.path.join(p, 'storeA'))
        self.print_log.info('Access to lustre database: {}'.format(ok))
        return ok
    
    @property
    def has_access_users_database(self):
        """Boolean specifying whether the users database can be accessed"""
        ok = False
        p = os.path.join(self.ROOTDIR, 'metno/')
        if os.path.exists(p):
            test_loc = os.path.join(p, 'aerocom-users-database')
            ok = self._check_access(test_loc)
                
        self.print_log.info("Access to aerocom-users-database: {}".format(ok))
        return ok
    
    @property
    def has_access_testdata(self):
        """Boolean specifying whether the testdataset can be accessed"""
        ok = False
        if self.dir_exists(self._testdatadir) and 'modeldata' in os.listdir(self.TESTDATADIR):
            ok = True
            
        self.print_log.info("Access to pyaerocom-testdata: {}".format(ok))
        return ok
    
    def _infer_config_file(self):
        """Infer the database configuration to be loaded"""
        if self.has_access_lustre:
            self.print_log.info("Init data paths for lustre")
            self.GRID_IO.load_aerocom_default()
            return self._config_files['metno']
        elif self.has_access_users_database:
            self.print_log.info("Init data paths for users database")
            self.GRID_IO.load_aerocom_default()
            return self._config_files['aerocom-users-database']
        elif self.has_access_testdata:
            self.print_log.info("Init data paths for pyaerocom testdata")
            self.GRID_IO.load_aerocom_default()
            return self._config_files['pyaerocom-testdata']
        
        self.GRID_IO.load_default()
        return None
    
    @property
    def ALL_DATABASE_IDS(self):
        '''ID's of available database configurations'''
        return list(self._config_files.keys())
    
    @property
    def ROOTDIR(self):
        return os.path.abspath(os.sep)
    
    @property
    def HOMEDIR(self):
        """Home directory of user"""
        return os.path.expanduser("~") + '/'
    
    @property
    def TESTDATADIR(self):
        return self._testdatadir
    
    @TESTDATADIR.setter
    def TESTDATADIR(self, val):
        if not self.dir_exists(val):
            raise ValueError('Cannot set pyaerocom-testdata directory {}.'
                             'Directory does not exist')
        self._testdatadir = val
        if not self.has_access_testdata:
            self._testdatadir = None
            raise IOError('Input path {} could not be identified as official '
                          'pyaerocom testdata directory (requires that a sub '
                          'directory modeldata exists)'.format(val))
        self.read_config(self._config_files['pyaerocom-testdata'], keep_basedirs=False)
    
    @property
    def OUTPUTDIR(self):
        """Default output directory"""
        return self._outputdir
    
    @property
    def COLOCATEDDATADIR(self):
        """Directory for accessing and saving colocated data objects"""
        return self._colocateddatadir
    
    @property
    def OUT_BASEDIR(self):
        msg = 'Attribute OUT_BASEDIR is deprecated. Please use OUTPUTDIR instead'
        self.print_log.warning(DeprecationWarning(msg))
        return self.OUTPUTDIR
    
    @property
    def CACHING(self):
        """Activate writing of and reading from cache files"""
        return self._caching_active
    
    @CACHING.setter
    def CACHING(self, val):
        self._caching_active = bool(val)
        
    @property
    def OBSDATACACHEDIR(self):
        """Cache directory for UngriddedData objects (deprecated)"""
        msg=('Attr. was renamed (but still works). Please us CACHEDIR instead')
        self.print_log.warning(DeprecationWarning(msg))
        return self.CACHEDIR
    
    @property
    def CACHEDIR(self):
        """Cache directory for UngriddedData objects"""
        if self._cachedir is None:
            raise IOError('Cache directory is not defined')
        try:
            return chk_make_subdir(self._cachedir, getpass.getuser())
        except Exception as e:
            from pyaerocom import print_log
            print_log.info('Failed to access CACHEDIR: {}\n'
                           'Deactivating caching'.format(repr(e)))
            self._caching_active = False
            
    @CACHEDIR.setter
    def CACHEDIR(self, val):
        """Cache directory"""
        if not os.path.exists(val):
            raise ValueError('Input directory does not exist {}'.format(val))
        elif not self._write_access(val):
            raise ValueError('Cannot write to {}'.format(val))
        self._cachedir = val
        
    @property
    def VAR_PARAM(self):
        """Deprecated name, please use :attr:`VARS` instead"""
        self.print_log.warning('Deprecated (but still functional) name '
                               'VARS. Please use VARS')
        return self.VARS
    
    @property
    def VARS(self):
        """Instance of class VarCollection (for default variable information)"""
        if self._var_param is None: #has not been accessed before
            self._var_param = VarCollection(self._var_info_file)
        return self._var_param
    
    @property
    def COORDINFO(self):
        if self._coords is None:
            self._coords = VarCollection(self._coords_info_file)
        return self._coords
    
    @property
    def LOGFILESDIR(self):
        """Directory where logfiles are stored"""
        try:
            logdir = chk_make_subdir(self.OUTPUTDIR, '_log')
            return logdir
        except Exception as e:
            from pyaerocom import print_log
            print_log.info('Failed to access LOGFILESDIR: {}'
                           'Deactivating file logging'.format(repr(e)))
            self.WRITE_FILEIO_ERR_LOG = False
            
       
    @property
    def MODELBASEDIR(self):
        """Base directory of model data
        
        If changed, all relevant subdirectories are updated as well.
        """
        return self._modelbasedir
    
    @MODELBASEDIR.setter
    def MODELBASEDIR(self, value):
        if not os.path.exists(value):
            raise IOError('Input directory does not exist')
        self._modelbasedir = value
        self.reload()
        self.check_data_dirs()
    
    @property
    def OBSBASEDIR(self):
        """Base directory of model data"""
        return self._obsbasedir
    
    @OBSBASEDIR.setter
    def OBSBASEDIR(self, value):
        if not os.path.exists(value):
            raise IOError('Input directory does not exist')
        self._obsbasedir = value
        self.reload()    
        self.check_data_dirs()
    
    @property 
    def BASEDIR(self):
        """Base directory of data
        
        Note
        ----
        If this attribute is changed it changes both, :attr:`MODELBASEDIR` and
        :attr:`OBSBASEDIR`.
        """
        return self._modelbasedir
    
    @BASEDIR.setter
    def BASEDIR(self, value):
        if not os.path.exists(value):
            raise IOError('Cannot change data base directory. Input directory '
                          'does not exist')
            
        self._obsbasedir = value
        self._modelbasedir = value
        
        subdirs = os.listdir(value)
        from pyaerocom import print_log
        if 'aerocom0' in subdirs:
            print_log.info('Initiating directories for lustre')
            self.read_config(self._config_ini, 
                             keep_basedirs=True)
        elif 'obsdata' in subdirs: #test dataset
            
            print_log.info('Initiating directories for pyaerocom testdataset')
            self.read_config(self._config_ini_testdata, 
                             keep_basedirs=True)
            self._cachedir = os.path.join('..', '_cache')
        elif 'AMAP' in subdirs:
            print_log.info('Initiating directories for AEROCOM users database')
            self.read_config(self._config_ini_user_server, 
                             keep_basedirs=True)
        else:
            self.reload()    
        
         
    @property
    def READY(self):
        """Checks if relevant directories exist, returns True or False"""
        return bool(self.check_directories())
    
    @property
    def DIR_INI_FILES(self):
        """Directory containing configuration files"""
        from pyaerocom import __dir__
        return os.path.join(__dir__, 'data')
    
    @property
    def EBASMC_SQL_DATABASE(self):
        """Path to EBAS SQL database"""
        return os.path.join(self.OBSCONFIG["EBASMC"]["PATH"], 
                                'ebas_file_index.sqlite3')
        
    @property
    def EBASMC_DATA_DIR(self):
        """Data directory of EBAS multicolumn files"""
        return os.path.join(self.OBSCONFIG["EBASMC"]["PATH"], 'data/')
    
    @property
    def EBAS_FLAGS_FILE(self):
        from pyaerocom import __dir__
        return os.path.join(__dir__, 'data', 'ebas_flags.csv')
    
    @property
    def OBSDIRS(self):
        """Direcories of observation networks"""
        return [x["PATH"] for x in self.OBSCONFIG.values()]
    
    @property 
    def OBS_START_YEARS(self):
        """Start years of observation networks"""
        return [x["START_YEAR"] for x in self.OBSCONFIG.values()]
    
    @property
    def OBS_IDS(self):
        """List of all IDs of observations"""
        return [x for x in self.OBSCONFIG.keys()]
    
    def make_default_vert_grid(self):
        """Makes default vertical grid for resampling of profile data"""
        step = self.DEFAULT_VERT_GRID_DEF['step']
        offs = int(step/2)
        return np.arange(self.DEFAULT_VERT_GRID_DEF['lower'] + offs,
                         self.DEFAULT_VERT_GRID_DEF['upper'] - offs,
                         step)
        
    def dir_exists(self, path):
        """Checks if directory exists"""
        if isinstance(path, str) and os.path.isdir(path):
            return True
        return False
    
    @staticmethod
    def _write_access(path):
        test = os.path.join(path, '_tmp')
        try:
            os.mkdir(test)
            os.rmdir(test)
            return True
        except:
            return False
        return os.access(path, os.W_OK)
    
    @staticmethod
    def _read_access(path):
        return os.access(path, os.R_OK)
    
    def check_directories(self):
        """Checks all predefined data directories for availability
        
        Prints each directory that is not available
        """
        self.logger.info('Checking data directories')
        ok =True
        #model_dirs = []
        # CHECK BASIC DATA READING DIRECTORIES
        if not self.dir_exists(self._modelbasedir):
            self.logger.warning("Model base directory {} does not exist"
                                .format(self._modelbasedir))
            ok=False
        if not self.dir_exists(self._obsbasedir):
            self.logger.warning("Observations base directory {} does not "
                                "exist".format(self._obsbasedir))
            ok=False
        
        return self.init_outputdirs() * ok
    
    def init_outputdirs(self):
        """Initiate output directories based on current configuration
        
        Checks, and if applicable, writes / creates required output directories
        (i.e. :attrs:`OUTPUTDIR, CACHEDIR, COLOCATEDDATADIR`).
        
        Returns
        -------
        bool
            True, if everything is okay, False if not
        """
        out_ok = True
        if not self.dir_exists(self._outputdir) or not self._write_access(self._outputdir):
            out_ok = False
            try:
                self._outputdir = chk_make_subdir(self.HOMEDIR, self._outhomename)
                out_ok = True
            except:
                self.logger.warning('Failed to create {} directory in home '
                                    'directory'.format(self._outhomename))
        
        if not out_ok or not self._write_access(self._outputdir):
            self.log.info('Cannot establish write access to output directory {}'
                           .format(self._outputdir))
            return False

        if (not self.dir_exists(self._colocateddatadir) or not 
                self._read_access(self._colocateddatadir)):
            self._colocateddatadir = os.path.join(self._outputdir,
                                                  'colocated_data')
        if not self.dir_exists(self._cachedir) or not self._write_access(self._cachedir):
            self._cachedir = chk_make_subdir(self._outputdir, '_cache')
        
        
        # if this file exists no cache file is read
        # used to ease debugging
        if self.dir_exists(self.CACHEDIR):
            self.DONOTCACHEFILE = os.path.join(self.CACHEDIR, 'DONOTCACHE')
            if os.path.exists(self.DONOTCACHEFILE):
                self._caching_active = False
        
        if not self._write_access(self._cachedir):
            self.logger.info('Cannot establish write access to cache '
                             'directory {}. Deactivating caching of files'
                             .format(self._cachedir))
            self._caching_active = False
        return out_ok
    
    def add_model_dir(self, dirname):
        """Add new model directory"""
        self.MODELDIRS.append(os.path.join(self.MODELBASEDIR, 'dirname'))
        
    def change_database(self, database_name='metno', keep_root=False):
        '''Changes the path setup for a specific data environment
        
        Parameters
        ----------
        database_name : str
            name of path environment for database. To see available database
            ID's use :attr:`ALL_DATABASE_IDS`
        keep_root : bool
            if True, :attr:`BASEDIR` remains unchanged and paths in
            corresponding ini files are set relative to current :attr:`BASEDIR`.
            Else, :attr:`BASEDIR` is updated using the specifications 
            provided in the corresponding ini file.
        '''
        if not database_name in self.ALL_DATABASE_IDS:
            raise ValueError('Unkown database name {}. Please choose from '
                             '{}'.format(database_name, self.ALL_DATABASE_IDS))
        self.read_config(self._config_files[database_name], 
                         keep_basedirs=keep_root)
        
    def reload(self, keep_basedirs=True):
        """Reload config file (for details see :func:`read_config`)"""
        self.read_config(self._config_ini, keep_basedirs)

    @property    
    def EBAS_FLAG_INFO(self):
        """Information about EBAS flags
        
        Dictionary containing 3 dictionaries (keys: ```valid, values, info```) 
        that contain information about validity of each flag (```valid```), 
        their actual values (```values```, e.g. V, M, I)
        """
        if self._ebas_flag_info is None:
            from pyaerocom.io.helpers import read_ebas_flags_file
            self._ebas_flag_info = read_ebas_flags_file(self.EBAS_FLAGS_FILE)
        return self._ebas_flag_info
    
    def load_default_config(self):
        """Read AeroCom default config file"""
        self.read_config(self._config_files['metno'])
        
    def read_config(self, config_file, keep_basedirs=True):
        """Read and import form paths.ini"""
        if not os.path.isfile(config_file):
            raise IOError("Configuration file paths.ini at %s does not exist "
                          "or is not a file"
                          %config_file)
        self.OBSCONFIG = od()
        cr = ConfigParser()
        cr.read(config_file)
        if cr.has_section('outputfolders'):
            if not keep_basedirs or not self.dir_exists(self._cachedir):
                try:
                    cachedir = cr['outputfolders']['CACHEDIR']
                    if not self._write_access(cachedir):
                        raise PermissionError('Cannot write to {}'.format(cachedir))
                    self._cachedir = cr['outputfolders']['CACHEDIR']
                except Exception as e:
                    self.logger.warning('Failed to init cache directory from '
                                        'config file. Error: {}'
                                        .format(repr(e)))
                
            if not keep_basedirs or not self.dir_exists(self._outputdir):
                try:
                    outdir = cr['outputfolders']['OUTPUTDIR']
                    if not self._write_access(outdir):
                        raise PermissionError('Cannot write to {}'.format(outdir))
                        
                    self._outputdir = outdir
                    self._colocateddatadir = os.path.join(outdir, 
                                                          'colocated_data')
                except Exception as e:
                    self.logger.warning('Failed to init output and colocated data '
                                        'directory from config file. Error: {}'
                                        .format(repr(e)))

            try:
                _dir = cr['outputfolders']['LOCALTMPDIR']
                # expand $HOME
                if '$HOME' in _dir:
                    _dir = _dir.replace('$HOME', os.path.expanduser('~'))
                if '${USER}' in _dir:
                    _dir = _dir.replace('${USER}', getpass.getuser())
                local_tmp_dir = _dir

                if not self._write_access(local_tmp_dir):
                    raise PermissionError('Cannot write to {}'.format(local_tmp_dir))
                self.LOCAL_TMP_DIR = local_tmp_dir

            except Exception as e:
                    self.logger.warning('Failed to init local tmp directory from '
                                        'config file. Error: {}'
                                        .format(repr(e)))
                    
        if cr.has_section('supplfolders'):
            for name, path in cr['supplfolders'].items():
                if not os.path.exists(path):
                    self.print_log.warning('Supplementary data directory for '
                                        '{} does not exist:\n{}'.format(name, path))
                self.SUPPLDIRS[name] = path
                
        #init base directories for Model data
        if not keep_basedirs or not self.dir_exists(self._modelbasedir):
            _dir = cr['modelfolders']['BASEDIR']
            if '$HOME' in _dir:
                _dir = _dir.replace('$HOME', os.path.expanduser('~'))
            self._modelbasedir = _dir
        
        self.MODELDIRS = (cr['modelfolders']['dir'].
                          replace('${BASEDIR}', self._modelbasedir).
                          replace('\n','').split(','))

        #Read directories for observation location
        if not keep_basedirs or not self.dir_exists(self._obsbasedir):
            _dir = cr['obsfolders']['BASEDIR']
            if '$HOME' in _dir:
                _dir = _dir.replace('$HOME', os.path.expanduser('~'))
            self._obsbasedir = _dir
        
        try:
            self._init_obsconfig(cr)
        except Exception as e:
            from pyaerocom import print_log
            print_log.exception('Failed to initiate obs config. Error: {}'
                                .format(repr(e)))
        cr.clear()
        self.check_directories()
    
    def _add_obsname(self, name):
        name_str = '{}_NAME'.format(name.upper())
        self[name_str] =  name
        return name_str
        
    def _add_obsnames_config(self, cr):
        names_cfg = []
        for obsname, ID in cr['obsnames'].items():
            name_str = '{}_NAME'.format(obsname.upper())
            self[name_str] =  ID
            names_cfg.append(name_str)
        return names_cfg
            
    def _init_obsconfig(self, cr):
        
        names_cfg = self._add_obsnames_config(cr)
        
        OBSCONFIG = self.OBSCONFIG
        for obsname, path in cr['obsfolders'].items():
            if obsname.lower() == 'basedir':
                continue
            name_str = '{}_NAME'.format(obsname.upper())
            if name_str in names_cfg:
                ID = self.__dict__[name_str]    
            else:
                ID = self._add_obsname(obsname)
            OBSCONFIG[ID] = {}
            p = path.replace('${BASEDIR}', self._obsbasedir)
            p = p.replace('$HOME', os.path.expanduser('~'))
            OBSCONFIG[ID]['PATH'] = p
            
        for obsname, year in cr['obsstartyears'].items():
            NAME = '{}_NAME'.format(obsname.upper())
            if NAME in self.__dict__:
                ID = self.__dict__[NAME]
                if ID in OBSCONFIG.keys():
                    OBSCONFIG[ID]['START_YEAR'] = year
        
        self.OBSCONFIG = OBSCONFIG
    
    def add_data_source(self, data_dir, name=None):
        """Add a network to the data search structure
        
        Parameters
        ----------
        name : str
            name of network 
        data_dir : str
            directory where data files are stored
        
        Raises
        ------
        AttributeError
            if the network name is already reserved 
        ValueError
            if the data directory does not exist
        """
        raise NotImplementedError('Coming soon... need some refactoring before')
        name_str = '{}_NAME'.format(name.upper())
        if name_str in self.__dict__.keys():
            raise AttributeError('Network with ID {} does already exist'.format(name_str))
        elif not os.path.exists(data_dir):
            raise ValueError('Input data directory does not exist')
        self[name_str] =  name
        self.OBSCONFIG[name] = {'PATH' : data_dir}
        
    def short_str(self):
        """Deprecated method"""
        return self.__str__()    
    
    def __setitem__(self, key, val):
        self.__dict__[key] = val
        
    def __str__(self):
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}\n".format(head, len(head)*"-")
        for k, v in self.__dict__.items():
            if k.startswith('_'):
                pass
            if k=='VARS':
                s += '\n{}\n{}'.format(k, list_to_shortstr(v.all_vars))
            elif isinstance(v, dict):
                s += "\n%s (dict)" %k
            elif isinstance(v, list):
                s += "\n%s (list)" %k
                s += list_to_shortstr(v)
            else:
                s += "\n%s: %s" %(k, v)
        return s
        
if __name__=="__main__":
    import pyaerocom as pya
    
    pya.const.COORDINFO.a
# =============================================================================
#     pya.const.BASEDIR = '/home/jonasg/aerocom-users-database'
#     
#     pya.browse_database('Aeronet*')
# =============================================================================



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Mathematical low level utility methods ofcd pyaerocom
"""

import numpy as np
from pyaerocom import const, logger
from scipy.stats import pearsonr, spearmanr, kendalltau

### LAMBDA FUNCTIONS
in_range = lambda x, low, high: low <= x <= high

### OTHER FUNCTIONS
def calc_statistics(data, ref_data, lowlim=None, highlim=None,
                    min_num_valid=5):
    """Calc statistical properties from two data arrays
    
    Calculates the following statistical properties based on the two provided
    1-dimensional data arrays and returns them in a dictionary (keys are 
    provided after the arrows):
        
        - Mean value of both arrays -> refdata_mean, data_mean
        - Standard deviation of both arrays -> refdata_std, data_std
        - RMS (Root mean square) -> rms
        - NMB (Normalised mean bias) -> nmb
        - MNMB (Modified normalised mean bias) -> mnmb
        - FGE (Fractional gross error) -> fge
        - R (Pearson correlation coefficient) -> R
        - R_spearman (Spearman corr. coeff) -> R_spearman
        
    Note
    ----
    Nans are removed from the input arrays, information about no. of removed
    points can be inferred from keys `totnum` and `num_valid` in return dict.
    
    Parameters
    ----------
    data : ndarray
        array containing data, that is supposed to be compared with reference
        data
    ref_data : ndarray
        array containing data, that is used to compare `data` array with
    lowlim : float
        lower end of considered value range (e.g. if set 0, then all datapoints
        where either ``data`` or ``ref_data`` is smaller than 0 are removed)
    highlim : float
        upper end of considered value range
    min_num_valid : int
        minimum number of valid measurements required to compute statistical
        parameters.
        
    Returns
    -------
    dict
        dictionary containing computed statistics    
    
    Raises
    ------
    ValueError 
        if either of the input arrays has dimension other than 1
        
    """
    data = np.asarray(data)
    ref_data = np.asarray(ref_data)
    
    if not data.ndim == 1 or not ref_data.ndim == 1:
        raise ValueError('Invalid input. Data arrays must be one dimensional')
        
    result = {}
    
    mask = ~np.isnan(ref_data) * ~np.isnan(data)
    num_points = mask.sum()
    
    data, ref_data = data[mask], ref_data[mask]
    
    result['totnum'] = float(len(mask))
    result['num_valid'] = float(num_points)
    result['refdata_mean'] = np.nanmean(ref_data)
    result['refdata_std'] = np.nanstd(ref_data)
    result['data_mean'] = np.nanmean(data)
    result['data_std'] = np.nanstd(data)
    if not num_points > min_num_valid:
        if lowlim is not None:
            valid = np.logical_and(data>lowlim, ref_data>lowlim)
            data = data[valid]
            ref_data = ref_data[valid]
        if highlim is not None:
            valid = np.logical_and(data<highlim, ref_data<highlim)
            data = data[valid]
            ref_data = ref_data[valid]
        
        result['rms'] = np.nan
        result['nmb'] = np.nan
        result['mnmb'] = np.nan
        result['fge'] = np.nan
        result['R'] = np.nan
        result['R_spearman'] = np.nan
    else:
        if lowlim is not None:
            valid = np.logical_and(data>lowlim, ref_data>lowlim)
            data = data[valid]
            ref_data = ref_data[valid]
        if highlim is not None:
            valid = np.logical_and(data<highlim, ref_data<highlim)
            data = data[valid]
            ref_data = ref_data[valid]
        
        difference = data - ref_data
        
        result['rms'] = np.sqrt(np.mean(difference**2))
        
        result['R'] = pearsonr(data, ref_data)[0]
        result['R_spearman'] = spearmanr(data, ref_data)[0]
        result['R_kendall'] = kendalltau(data, ref_data)[0]
        
        # NMB, MNMB and FGE are constrained to positive values, thus negative
        # values need to be removed
        neg_ref = ref_data < 0
        neg_data = data < 0
        
        use_indices = ~(neg_data + neg_ref)
        
        diff_pos = difference[use_indices]
        ref_data_pos = ref_data[use_indices]
        data_pos = data[use_indices]
        
        num_points_pos = len(data_pos)
        
        if num_points_pos == 0:
            result['nmb'] = np.nan
            result['mnmb'] = np.nan
            result['fge'] = np.nan
        else:
            result['nmb'] = np.sum(diff_pos) / np.sum(ref_data_pos) #*100.
            
            tmp = diff_pos / (data_pos + ref_data_pos)
            
            result['mnmb'] = 2. / num_points_pos * np.sum(tmp)# * 100.
            result['fge'] = 2. / num_points_pos * np.sum(np.abs(tmp)) #* 100.
            
        result['num_neg_data'] = np.sum(neg_data)
        result['num_neg_refdata'] = np.sum(neg_ref)
     
    return result

def closest_index(num_array, value):
    """Returns index in number array that is closest to input value"""
    return np.argmin(np.abs(np.asarray(num_array) - value))

def numbers_in_str(input_string):
    """This method finds all numbers in a string
    
    Note
    ----
    - Beta version, please use with care
    - Detects only integer numbers, dots are ignored
    
    Parameters
    ----------
    input_string : str
        string containing numbers
    
    Returns
    -------
    list
        list of strings specifying all numbers detected in string
        
    Example
    -------
    >>> numbers_in_str('Bla42Blub100')
    [42, 100]
    """
    numbers = []
    IN_NUM=False
    c_num = None
    for char in input_string:
        try:
            int(char)
            if not IN_NUM:
                IN_NUM = True
                c_num = char
            elif IN_NUM:
                c_num += char
        except:
            if IN_NUM:
                numbers.append(c_num)
            IN_NUM=False
    if IN_NUM:
        numbers.append(c_num)
    return numbers    
    
def calc_ang4487aer(data):
    """Compute Angstrom coefficient (440-870nm) from 440 and 870 nm AODs
    
    Parameters
    ----------
    data : dict-like
        data object containing imported results
    
    Note
    ----
    Requires the following two variables to be available in provided data 
    object:
        
        1. od440aer
        2. od870aer
    
    Returns
    -------
    ndarray
        array containing computed angstrom coefficients
    
    Raises 
    ------
    AttributError
        if either 'od440aer' or 'od870aer' are not available in data object
    """
    if not all([x in data for x in ['od440aer','od870aer']]):
        raise AttributeError("Either of the two (or both) required variables "
                             "(od440aer, od870aer) are not available in data")
    od440aer, od870aer = data['od440aer'], data['od870aer']
    return compute_angstrom_coeff(od440aer, od870aer, .44, .87)

def calc_od550aer(data):
    """Compute AOD at 550 nm using Angstrom coefficient and 500 nm AOD
        
    Parameters
    ----------
    data : dict-like
        data object containing imported results
    
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(data=data, 
                           var_name='od550aer', 
                           to_lambda=.55, 
                           od_ref='od500aer', 
                           lambda_ref=.50, 
                           od_ref_alt='od440aer', 
                           lambda_ref_alt=.44,
                           use_angstrom_coeff='ang4487aer')

def calc_abs550aer(data):
    """Compute AOD at 550 nm using Angstrom coefficient and 500 nm AOD
        
    Parameters
    ----------
    data : dict-like
        data object containing imported results
    
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(data=data, 
                           var_name='abs550aer', 
                           to_lambda=.55, 
                           od_ref='abs500aer', 
                           lambda_ref=.50, 
                           od_ref_alt='abs440aer', 
                           lambda_ref_alt=.44,
                           use_angstrom_coeff='angabs4487aer')
    
def calc_od550gt1aer(data):
    """Compute coarse mode AOD at 550 nm using Angstrom coeff. and 500 nm AOD
        
    Parameters
    ----------
    data : dict-like
        data object containing imported results
    
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(data=data, 
                           var_name='od550gt1aer', 
                           to_lambda=.55, 
                           od_ref='od500gt1aer', 
                           lambda_ref=.50,
                           use_angstrom_coeff='ang4487aer')

def calc_od550lt1aer(data):
    """Compute fine mode AOD at 550 nm using Angstrom coeff. and 500 nm AOD
        
    Parameters
    ----------
    data : dict-like
        data object containing imported results
    
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    """
    return _calc_od_helper(data=data, 
                           var_name='od550lt1aer', 
                           to_lambda=.55, 
                           od_ref='od500lt1aer', 
                           lambda_ref=.50,
                           use_angstrom_coeff='ang4487aer')
        
    
def compute_angstrom_coeff(od1, od2, lambda1, lambda2):
    """Compute Angstrom coefficient based on 2 optical densities
    
    Parameters
    ----------
    od1 : :obj:`float` or :obj:`ndarray`
        AOD at wavelength 1
    od2 : :obj:`float` or :obj:`ndarray`
        AOD at wavelength 2
    lambda1 : :obj:`float` or :obj:`ndarray`
        wavelength 1
    lambda 2 : :obj:`float` or :obj:`ndarray`
        wavelength 2
        
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        Angstrom exponent(s)
    """
    return -np.log(od1 / od2) / np.log(lambda1 / lambda2)

def compute_od_from_angstromexp(to_lambda, od_ref, lambda_ref, 
                                 angstrom_coeff):
    """Compute AOD at specified wavelength 
    
    Uses Angstrom coefficient and reference AOD to compute the 
    corresponding wavelength shifted AOD
    
    Parameters
    ----------
    to_lambda : :obj:`float` or :obj:`ndarray`
        wavelength for which AOD is calculated
    od_ref : :obj:`float` or :obj:`ndarray`
        reference AOD
    lambda_ref : :obj:`float` or :obj:`ndarray`
        wavelength corresponding to reference AOD
    angstrom_coeff : :obj:`float` or :obj:`ndarray`
        Angstrom coefficient
        
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
    
    """
    return od_ref * (lambda_ref / to_lambda) ** angstrom_coeff

def _calc_od_helper(data, var_name, to_lambda, od_ref, lambda_ref, 
                    od_ref_alt=None, lambda_ref_alt=None, 
                    use_angstrom_coeff='ang4487aer'):
    """Helper method for computing ODs
    
    Parameters
    ----------
    data : dict-like
        data object containing loaded results used to compute the ODs at a new
        wavelength
    var_name : str
        name of variable that is supposed to be computed (is used in order to 
        see whether a global lower threshold is defined for this variable and
        if this is the case, all computed values that are below this threshold
        are replaced with NaNs)
    to_lambda : float
        wavelength of computed AOD
    od_ref : :obj:`float` or :obj:`ndarray`
        reference AOD
    lambda_ref : :obj:`float` or :obj:`ndarray`
        wavelength corresponding to reference AOD
    od_ref_alt : :obj:`float` or :obj:`ndarray`, optional
        alternative reference AOD (is used for datapoints where former is 
        invalid)
    lambda_ref_alt : :obj:`float` or :obj:`ndarray`, optional
        wavelength corresponding to alternative reference AOD
    use_angstrom_coeff : str
        name of Angstrom coefficient in data, that is used for computation
        
    Returns
    -------
    :obj:`float` or :obj:`ndarray`
        AOD(s) at shifted wavelength
        
    Raises
    ------
    AttributeError
        if neither ``od_ref`` nor ``od_ref_alt`` are available in data, or if
        ``use_angstrom_coeff`` is missing
    """
    if not od_ref in data:
        logger.warning('Reference OD at {} nm is not available in data, '
                       'checking alternative'.format(lambda_ref))
        if od_ref_alt is None or not od_ref_alt in data:
            raise AttributeError('No alternative OD found for computation of '
                                 '{}'.format(var_name))
        return compute_od_from_angstromexp(to_lambda=to_lambda, 
                                           od_ref=data[od_ref_alt],
                                           lambda_ref=lambda_ref_alt, 
                                           angstrom_coeff=data[use_angstrom_coeff])
    elif not use_angstrom_coeff in data:
        raise AttributeError("Angstrom coefficient (440-870 nm) is not "
                             "available in provided data")
    result = compute_od_from_angstromexp(to_lambda=to_lambda, 
                                          od_ref=data[od_ref],
                                          lambda_ref=lambda_ref, 
                                          angstrom_coeff=data[use_angstrom_coeff])
    # optional if available
    if od_ref_alt in data:
        # fill up time steps that are nans with values calculated from the
        # alternative wavelength to minimise gaps in the time series
        mask = np.argwhere(np.isnan(result))
        
        if len(mask) > 0: #there are nans
            ods_alt = data[od_ref_alt][mask]
            ang = data[use_angstrom_coeff][mask]
            replace = compute_od_from_angstromexp(to_lambda=to_lambda, 
                                                    od_ref=ods_alt,
                                                    lambda_ref=lambda_ref_alt, 
                                                    angstrom_coeff=ang)
            result[mask] = replace
    
    try:
        # now replace all values with NaNs that are below the global lower threshold
        below_thresh = result < const.VARS[var_name]['minimum']
        result[below_thresh] = np.nan
    except:
        logger.warning("Could not access lower limit from global settings for "
                       "variable {}".format(var_name))
    
    return result

def compute_ang4470dryaer_from_dry_scat(data):
    """Compute angstrom exponent between 440 and 700 nm 
    
    Parameters
    ----------
    StationData or dict
        data containing dry scattering coefficients at 440 and 700 nm
        (i.e. keys scatc440dryaer and scatc700dryaer)
    
    Returns
    -------
    StationData or dict
        extended data object containing angstrom exponent
    """
    return compute_angstrom_coeff(data['scatc440dryaer'],
                                  data['scatc700dryaer'],
                                  440, 700)

def compute_scatc550dryaer(data):
    """Compute dry scattering coefficent applying RH threshold
    
    Cf. :func:`_compute_dry_helper`
    
    Parameters
    ----------
    dict
        data object containing scattering and RH data
    
    Returns
    -------
    dict 
        modified data object containing new column scatc550dryaer
    
    """
    rh_max= const.VARS['scatc550dryaer'].dry_rh_max
    return _compute_dry_helper(data, data_colname='scatc550aer', 
                               rh_colname='scatcrh', 
                               rh_max_percent=rh_max)
    
def compute_scatc440dryaer(data):
    """Compute dry scattering coefficent applying RH threshold
    
    Cf. :func:`_compute_dry_helper`
    
    Parameters
    ----------
    dict
        data object containing scattering and RH data
    
    Returns
    -------
    dict 
        modified data object containing new column scatc550dryaer
    
    """
    rh_max= const.VARS['scatc440dryaer'].dry_rh_max
    return _compute_dry_helper(data, data_colname='scatc440aer', 
                               rh_colname='scatcrh', 
                               rh_max_percent=rh_max)

def compute_scatc700dryaer(data):
    """Compute dry scattering coefficent applying RH threshold
    
    Cf. :func:`_compute_dry_helper`
    
    Parameters
    ----------
    dict
        data object containing scattering and RH data
    
    Returns
    -------
    dict 
        modified data object containing new column scatc550dryaer
    
    """
    rh_max= const.VARS['scatc700dryaer'].dry_rh_max
    return _compute_dry_helper(data, data_colname='scatc700aer', 
                               rh_colname='scatcrh', 
                               rh_max_percent=rh_max)
    
def compute_absc550dryaer(data):
    """Compute aerosol dry absorption coefficent applying RH threshold
    
    Cf. :func:`_compute_dry_helper`
    
    Parameters
    ----------
    dict
        data object containing scattering and RH data
    
    Returns
    -------
    dict 
        modified data object containing new column scatc550dryaer
    
    """
    rh_max= const.VARS['absc550dryaer'].dry_rh_max
    return _compute_dry_helper(data, data_colname='absc550aer', 
                               rh_colname='abscrh', 
                               rh_max_percent=rh_max)
    
def _compute_dry_helper(data, data_colname, rh_colname, 
                        rh_max_percent=None):
    """Compute new column that contains data where RH is smaller than ...
    
    All values in original data columns are set to NaN, where RH exceeds a 
    certain threshold or where RH is NaN.
    
    Parameters
    ----------
    data : dict-like
        dictionary-like object that contains data
    data_colname : str
        column name of variable data that is supposed to be filtered
    rh_colname : str
        column name of RH data
    rh_max_percent : int
        maximum relative humidity
        
    Returns
    -------
    dict
        modified data dictionary with new dry data column 
    """
    if rh_max_percent is None:
        rh_max_percent = const.RH_MAX_PERCENT_DRY
        
    vals = np.array(data[data_colname], copy=True)

    rh = data[rh_colname]
    
    high_rh = rh > rh_max_percent
    
    vals[high_rh] = np.nan
    vals[np.isnan(rh)] = np.nan
    
    return vals
    
def exponent(num):
    """Get exponent of input number
        
    Parameters
    ----------
    num : :obj:`float` or iterable
        input number
    
    Returns
    -------
    :obj:`int` or :obj:`ndarray` containing ints
        exponent of input number(s)
        
    Example
    -------
    >>> from pyaerocom.mathutils import exponent
    >>> exponent(2340)
    3
    """
    return np.floor(np.log10(abs(np.asarray(num)))).astype(int)

def range_magnitude(low, high):
    """Returns magnitude of value range
    
    Parameters
    ----------
    low : float
        lower end of range
    high : float
        upper end of range
    
    Returns
    -------
    int
        magnitudes spanned by input numbers
    
    Example
    -------
    
    >>> range_magnitude(0.1, 100)
    3
    >>> range_magnitude(100, 0.1)
    -3
    >>> range_magnitude(1e-3, 1e6)
    9
    
    """
    return exponent(high) - exponent(low)

if __name__ == "__main__":
    #import doctest
    exp = exponent(23)
    print(numbers_in_str('Bla42Blub100'))
    #run tests in all docstrings
    #doctest.testmod()
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 16 09:03:31 2018

@author: jonasg
"""
from pyaerocom import logger, print_log
from pyaerocom._lowlevel_helpers import BrowseDict
from pyaerocom.griddeddata import GriddedData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.colocateddata import ColocatedData
from pyaerocom.region import Region

class Filter(BrowseDict):
    """Class that can be used to filter gridded and ungridded data objects
    
    Note
    ----
    BETA version (currently being tested)
    
    Todo
    ----
    Include also temporal filtering and other filter options (e.g. variable, 
    etc.)
    
    Attributes
    ----------
    lon_range : list
        2-element list or array specifying longitude range
    lat_range : list
        2-element list or array specifying latitude range
    alt_range : list
        2-element list or array specifying altitude range
        
    Example
    -------
    >>> import pyaerocom as pya
    >>> data = pya.io.ReadGridded('ECMWF_OSUITE').read_var('od550aer')
    >>> data
    pyaerocom.GriddedData
    Grid data: <iris 'Cube' of Aerosol optical depth at 550 nm / (1) (time: 3287; latitude: 161; longitude: 320)>
    >>> regfilter = pya.Filter('EUROPE-noMOUNTAINS')
    >>> data_filtered = regfilter(data)
    >>> data_filtered
    pyaerocom.GriddedData
    Grid data: <iris 'Cube' of Aerosol optical depth at 550 nm / (1) (time: 3287; latitude: 45; longitude: 80)>
    """
    #: dictionary specifying altitude filters
    ALTITUDE_FILTERS = {'wMOUNTAINS'    :   None, #reserve namespace for
                        'noMOUNTAINS'   :   [-1e6, 1e3]} # 1000 m upper limit
    NO_FILTER_NAME = 'WORLD-wMOUNTAINS'
    def __init__(self, name=None, region=None, altitude_filter=None, **kwargs):
        # default name (i.e. corresponds to no filtering)
        self._name = self.NO_FILTER_NAME
        self._region = None
        
        self.lon_range = None
        self.lat_range = None
        self.alt_range = None
        
        if name is not None:
            self.infer_from_name(name)
        else:
            self.infer_from_name(self._name)
            
        if region is not None:
            self.set_region(region)
        if altitude_filter is not None:
            self.set_altitude_filter(altitude_filter)
        
        self.update(**kwargs)
        
    def infer_from_name(self, name):
        """Infer filter from name string
        
        Parameters
        ----------
        name : str
            name string in Aerocom format (e.g. WORLD-wMOUNTAINS)
        
        Raises
        ------
        IOError
            if region and altitude filter cannot be inferred
        """
        if not isinstance(name, str):
            raise IOError('Invalid input for name, need string, got {}'.format(type(name)))
        spl = name.split('-')
        # intitialise
        self.set_region(spl[0])
        if len(spl) > 1:
            alt_filter = spl[1]
        else:
            alt_filter = 'wMOUNTAINS'
        self.set_altitude_filter(alt_filter)
    
    def set_altitude_filter(self, filter_name):
        """Set default altitude filter"""
        if not filter_name in self.ALTITUDE_FILTERS:
            raise AttributeError('No such altitude filter: {}. Available '
                                 'filters are: {}'.format(filter_name, 
                                               self.ALTITUDE_FILTERS.keys()))
        self.alt_range = self.ALTITUDE_FILTERS[filter_name]
        
        spl = self.name.split('-')
        self._name = '{}-{}'.format(spl[0], filter_name)
    
    def set_region(self, region):
        if isinstance(region, str):
            region = Region(region)
        if not isinstance(region, Region):
            raise IOError('Invalid input for region, need string or '
                          'instance of Region class, got {}'.format(region))
        self.lon_range = region.lon_range
        self.lat_range = region.lat_range
        self._region = region
        
        spl = self.name.split('-')
        self._name = '{}-{}'.format(region.name, spl[1])
    
    @property
    def region_name(self):
        return self._region.name
    
    @property
    def name(self):
        return self._name
    
    def to_dict(self):
        """Convert filter to dictionary"""
        return {'region'    :   self.region_name, 
                'lon_range' :   self.lon_range,
                'lat_range' :   self.lat_range,
                'alt_range' :   self.alt_range}
        
    def _apply_ungridded(self, data_obj):
        """Apply filter to instance of class :class:`UngriddedData`
        """
        return data_obj.filter_by_meta(longitude=self.lon_range,
                                       latitude=self.lat_range,
                                       altitude=self.alt_range)
    
    def _apply_gridded(self, data_obj):
        """Apply filter to instance of class :class:`GriddedData`
        """
        print_log.warning('Applying regional cropping in GriddedData using Filter '
                       'class. Note that this does not yet include potential '
                       'cropping in the vertical dimension. Coming soon...')
        return data_obj.crop(region=self._region)
    
    def _apply_colocated(self, data_obj):
        print_log.warning('Applying regional cropping in ColocatedData using Filter '
                       'class. Note that this does not yet include potential '
                       'cropping in the vertical dimension. Coming soon...')
        return data_obj.apply_latlon_filter(region_id=self.region_name)
        
    def apply(self, data_obj):
        """Apply filter to data object
        
        Parameters
        ----------
        data_obj : :obj:`UngriddedData`, :obj:`GriddedData`
            input data object that is supposed to be filtered
            
        Returns
        -------
        :obj:`UngriddedData`, :obj:`GriddedData`
            filtered data object
            
        Raises
        ------
        IOError
            if input is invalid
        """
        if self.name == self.NO_FILTER_NAME:
            logger.info('NO FILTER flag: {} -> no filtering will be applied '
                        'in {}. Returning unchanged object.'
                        .format(self.NO_FILTER_NAME, type(data_obj)))
            return data_obj
        if isinstance(data_obj, UngriddedData):
            return self._apply_ungridded(data_obj)
        elif isinstance(data_obj, GriddedData):
            return self._apply_gridded(data_obj)
        elif isinstance(data_obj, ColocatedData):
            return self._apply_colocated(data_obj)
        raise IOError('Cannot filter {} obj, need instance of GriddedData or '
                'UngriddedData'.format(type(data_obj)))
      
    def __call__(self, data_obj):
        return self.apply(data_obj)
    
    
if __name__=="__main__":
    f = Filter('EUROPE-wMOUNTAINS')
    print(f)
    f.set_region('NAMERICA')
    print(f)  
    f.set_altitude_filter('noMOUNTAINS')     
    print(f.to_dict())
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
                    
def main(): 
    from argparse import ArgumentParser
    import os, sys

    # Define command line interface
    p = ArgumentParser(description='pyaerocom command line interface for '
                       'processing of trends data for Aerosol trends '
                       'interface')    
    # Mandatoory: name of configuration (should be in config file name)
    p.add_argument('name', help='Project ID')
    p.add_argument('-d', '--config_dir', default='.', 
                   help=('Directory containing trends config files'))
    
    p.add_argument('-o', '--obs_name', default=None, 
                   help=('Name of observation network that is supposed to be '
                         'run (default is all obsnetworks that are specified in '
                         'the config file'))
    p.add_argument('-c', '--clear', action='store_true',
                   help='Delete (potentially) existing json files before rerun')
    
    
    p.add_argument('-p', '--print_config', action='store_true',
                   help='print configuration (no analysis is run')
    
    
    
    args = p.parse_args()
    
    if args.clear:
        clear_existing_json = True
    else:
        clear_existing_json = None
        
    if not os.path.exists(args.config_dir):
        raise FileNotFoundError('No such file or directory: {}'.format(args.config_dir))
    
    import pyaerocom as pya
    cfg_avail = pya.web.helpers.get_all_config_files_trends_iface(args.config_dir)
    
    if len(cfg_avail) == 0:
        raise FileNotFoundError('Could not find any valid configuration files '
                                'in {}'.format(os.path.abspath(args.config_dir)))
    
    if not args.name in cfg_avail:
        raise ValueError('No such configuration with name {} available. '
                         'Please choose from {}'
                         .format(args.name, cfg_avail.keys()))
        
    te = pya.web.TrendsEvaluation(config_file=cfg_avail[args.name])
    
    if args.print_config:
        print(te)
        sys.exit()
    
    if args.obs_name is not None:
        if not args.obs_name in list(te.obs_config):
            raise ValueError('No such obs_name available in current config. '
                             'Please choose from: {}'
                             .format(list(te.obs_config)))

    te.run_evaluation(obs_name=args.obs_name, 
                      clear_existing_json=clear_existing_json)
    
if __name__ == '__main__':    
    main()
    


# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as file:
    readme = file.read()

with open('VERSION.md') as f:
    version = f.readline()
    print(version)
    f.close()


setup(
    name        =   'pyaerocom',
    version     =   version,
    author      =   'Jonas Gliss',
    author_email=   'jonas.gliss@met.no',
    url         =   'https://github.com/jgliss/pyaerocom.git',
    license     =   'GPLv3',
    package_dir =   {'pyaerocom'      :   'pyaerocom'},
    packages    =   ['pyaerocom',
                     'pyaerocom.io',
                     'pyaerocom.plot',
                     'pyaerocom.web',
                     'pyaerocom.interactive',
                     'pyaerocom.web.cli',
                     'pyaerocom.test',
                     'pyaerocom.io.test'],
    #include_package_data = True,            
    package_data=   {'pyaerocom'    :   ['data/*']},

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',

        # Pick your license as you wish (should match 'license' above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.,
        'Programming Language :: Python :: 3.6'
    ],

    install_requires    =   [],
    dependency_links    =   [],
    description = ('Python tools for the Aerocom project'),
    long_description = readme,
    entry_points = {'console_scripts' : [
            'pyaeroeval=pyaerocom.web.cli.main_aerocom_evaluation:main',
            'pyaerotrends=pyaerocom.web.cli.main_trends_evaluation:main'
            ]},
    zip_safe = False
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import xarray as xray
from pyaerocom import VerticalProfile, logger, const

from pyaerocom.exceptions import (MetaDataError, VarNotAvailableError,
                                  DataExtractionError, DataDimensionError,
                                  UnitConversionError, DataUnitError,
                                  TemporalResolutionError)
from pyaerocom._lowlevel_helpers import (dict_to_str, list_to_shortstr, 
                                         BrowseDict)
from pyaerocom.metastandards import StationMetaData
from pyaerocom.tstype import TsType
from pyaerocom.time_resampler import TimeResampler
from pyaerocom.helpers import isnumeric, isrange

from pyaerocom.units_helpers import convert_unit, unit_conversion_fac
from pyaerocom.time_config import PANDAS_FREQ_TO_TS_TYPE

class StationData(StationMetaData):
    """Dict-like base class for single station data
    
    ToDo: write more detailed introduction

    Note
    ----
    Variable data (e.g. numpy array or pandas Series) can be directly
    assigned to the object. When  assigning variable data it is
    recommended to add variable metadata (e.g. unit, ts_type)
    in :attr:`var_info`, where key is variable name and value is dict with
    metadata entries.

    Attributes
    ----------
    dtime : list
        list / array containing time index values
    var_info : dict
        dictionary containing information about each variable
    data_err : dict
        dictionary that may be used to store uncertainty timeseries or data
        arrays associated with the different variable data.
    overlap : dict
        dictionary that may be filled to store overlapping timeseries data 
        associated with one variable. This is, for instance, used in 
        :func:`merge_vardata` to store overlapping data from another station.
    
    """
    #: List of keys that specify standard metadata attribute names. This 
    #: is used e.g. in :func:`get_meta`
    STANDARD_COORD_KEYS = const.STANDARD_COORD_NAMES
    
    #: maximum numerical distance between coordinates associated with this
    #: station
    _COORD_MAX_VAR = 0.1 #km
    STANDARD_META_KEYS = list(StationMetaData().keys())
    
    VALID_TS_TYPES = const.GRID_IO.TS_TYPES
    
    def __init__(self, **meta_info):

        self.dtime = []
    
        self.var_info = BrowseDict()
        
        self.station_coords = dict.fromkeys(self.STANDARD_COORD_KEYS)
        
        self.data_err = BrowseDict()        
        self.overlap = BrowseDict()
        self.data_flagged = BrowseDict()
        
        super(StationData, self).__init__(**meta_info)
    
    @property
    def default_vert_grid(self):
        """AeroCom default grid for vertical regridding
        
        For details, see :attr:`DEFAULT_VERT_GRID_DEF` in :class:`Config`
        
        Returns
        -------
        ndarray
            numpy array specifying default coordinates
        """
        return const.make_default_vert_grid()
    
    def has_var(self, var_name):
        """Checks if input variable is available in data object

        Parameters
        ----------
        var_name : str
            name of variable

        Returns
        -------
        bool
            True, if variable data is available, else False
        """
        if not var_name in self:
            return False
        if not var_name in self.var_info:
            const.print_log.warning('Variable {} exists in data but has no '
                                    'metadata assigned in :attr:`var_info`'
                                    .format(var_name))
        return True

    def get_unit(self, var_name):
        """Get unit of variable data
        
        Parameters
        ----------
        var_name : str
            name of variable
        
        Returns
        -------
        str
            unit of variable
        
        Raises
        ------
        MetaDataError
            if unit cannot be accessed for variable
        """
        if not var_name in self.var_info:
            raise MetaDataError('Could not access variable metadata dict '
                                'for {}.'.format(var_name))
        try:
            return str(self.var_info[var_name]['units'])
        except KeyError:
            add_str = ''
            if 'unit' in self.var_info[var_name]:
                add_str = ('Corresponding var_info dict contains '
                           'attr. "unit", which is deprecated, please '
                           'check corresponding reading routine. ')
            raise MetaDataError('Failed to access units attribute for variable '
                                '{}. {}'.format(var_name, add_str))

    @property
    def units(self):
        """Dictionary containing units of all variables in this object"""
        ud = {}
        for var in self.var_info:
            ud[var] = self.get_unit(var)
        return ud

    def compute_trend(self, var_name, start_year=None, stop_year=None,
                      season=None, slope_confidence=None, **alt_range):
        if not self.has_var(var_name):
            raise VarNotAvailableError('No such variables {} in StationData'
                                       .format(var_name))
        from pyaerocom.trends_helpers import compute_trends_station
        return compute_trends_station(self, var_name, start_year, stop_year, 
                                      season, slope_confidence, **alt_range)
        
    def check_var_unit_aerocom(self, var_name):
        """Check if unit of input variable is AeroCom default, if not, convert

        Parameters
        ----------
        var_name : str
            name of variable

        Raises
        ------
        MetaDataError
            if unit information is not accessible for input variable name
        UnitConversionError
            if current unit cannot be converted into specified unit
            (e.g. 1 vs m-1)
        DataUnitError
            if current unit is not equal to AeroCom default and cannot
            be converted.
        """
        to_unit = const.VARS[var_name].units
        try:
            self.check_unit(var_name, to_unit)
        except:
            try:
                self.convert_unit(var_name, to_unit)
            except UnitConversionError as e:
                raise UnitConversionError('Failed to convert unit of variable '
                                          '{}. Reason: {}'
                                          .format(var_name, repr(e)))

    def check_unit(self, var_name, unit=None):
        """Check if variable unit corresponds to a certain unit
        
        Parameters
        ----------
        var_name : str
            variable name for which unit is to be checked
        unit : :obj:`str`, optional
            unit to be checked, if None, AeroCom default unit is used
            
        Raises
        ------
        MetaDataError
            if unit information is not accessible for input variable name
        UnitConversionError
            if current unit cannot be converted into specified unit 
            (e.g. 1 vs m-1)
        DataUnitError
            if current unit is not equal to input unit but can be converted 
            (e.g. 1/Mm vs 1/m)
        """
        if unit is None:
            unit = const.VARS[var_name].units
        u = self.get_unit(var_name)
        if not unit_conversion_fac(u, unit) == 1:
            raise DataUnitError('Invalid unit {} (expected {})'
                                .format(u, unit))
        
      
    def convert_unit(self, var_name, to_unit):
        """Try to convert unit of data
        
        Requires that unit of input variable is available in :attr:`var_info` 
        
        Note
        ----
        BETA version
        
        Parameters
        ----------
        var_name : str
            name of variable
        to_unit : str
            new unit
        
        Raises
        ------
        MetaDataError
            if variable unit cannot be accessed
        UnitConversionError
            if conversion failed
        """
        unit = self.get_unit(var_name)
        
        data = self[var_name]
        data = convert_unit(data, from_unit=unit, to_unit=to_unit, 
                            var_name=var_name)
        #conv_fac = unit_conversion_fac(unit, to_unit)
        
        #data *= conv_fac
        self[var_name] = data
        self.var_info[var_name]['units'] = to_unit
        const.logger.info('Successfully converted unit of variable {} in {} '
                          'from {} to {}'.format(var_name, self.station_name,
                                                 unit, to_unit))
    
        
    def dist_other(self, other):
        """Distance to other station in km
        
        Parameters
        ----------
        other : StationData
            other data object
            
        Returns
        -------
        float
            distance between this and other station in km
        """
        from pyaerocom.geodesy import calc_distance
        
        cthis = self.get_station_coords()
        cother = other.get_station_coords()
        
        return calc_distance(cthis['latitude'], cthis['longitude'],
                             cother['latitude'], cother['longitude'],
                             cthis['altitude'], cother['altitude'])
        
    def same_coords(self, other, tol_km=None):
        """Compare station coordinates of other station with this station
        
        Paremeters
        ----------
        other : StationData
            other data object
        tol_km : float
            distance tolerance in km
            
        Returns
        -------
        bool
            if True, then the two object are located within the specified 
            tolerance range
        """
        if tol_km is None:
            tol_km = self._COORD_MAX_VAR
        return True if self.dist_other(other) < tol_km else False
    
    def get_station_coords(self, force_single_value=True, quality_check=True):
        """Return coordinates as dictionary
        
        This method uses the standard coordinate names defined in 
        :attr:`STANDARD_COORD_KEYS` (latitude, longitude and altitude) to get
        the station coordinates. For each of these parameters tt first looks 
        in :attr:`station_coords` if the parameter is defined (i.e. it is not 
        None) and if not it checks if this object has an attribute that has 
        this name and uses that one. 
        
        Parameters
        ----------
        force_single_value : bool
            if True and coordinate values are lists or arrays, then they are 
            collapsed to single value using mean
        quality_check : bool
            if True, and coordinate values are lists or arrays, then the 
            standarad deviation in the values is compared to the upper limits
            allowed in the local variation. 
        
        Returns
        -------
        dict
            dictionary containing the retrieved coordinates
            
        Raises
        ------
        AttributeError
            if one of the coordinate values is invalid
        CoordinateError
            if local variation in either of the three spatial coordinates is
            found too large
        """
        _check_var = False
        vals , stds = {}, {}
        for key in self.STANDARD_COORD_KEYS:
            # prefer explicit if defined in station_coord dictionary (e.g. altitude
            # attribute in lidar data will be an array corresponding to profile
            # altitudes)
            val = self.station_coords[key]
            if val is not None:
                if not isnumeric(val):
                    raise MetaDataError('Station coordinate {} must be numeric. '
                                        'Got: {}'.format(key, val))
                vals[key] = val
                stds[key] = 0
            else:
                if not key in self or self[key] is None:
                    raise MetaDataError('{} information is not available in data'.format(key))
                val = self[key]
                std = 0
                # TODO: review the quality check and make shorter
                if force_single_value and not isinstance(val, (float, np.floating)):
                    if isinstance(val, (int, np.integer)):
                        val = np.float64(val)
                    elif isinstance(val, (list, np.ndarray)):
                        val = np.mean(val)
                        std = np.std(val)
                        if std > 0:
                            _check_var = True
                    else:
                        raise AttributeError("Invalid value encountered for coord "
                                             "{}, need float, int, list or ndarray, "
                                             "got {}".format(key, type(val)))
                vals[key] = val
                stds[key] = std
        if _check_var:
            raise NotImplementedError('This feature does currently not work '
                                      'due to recent API changes')
# =============================================================================
#             logger.debug("Performing quality check for coordinates")
#             lat, dlat, dlon, dalt = (vals['latitude'],
#                                      stds['latitude'],
#                                      stds['longitude'],
#                                      stds['altitude'])
#             lat_len = 111e3 #approximate length of latitude degree in m
#             if self.COORD_MAX_VAR['latitude'] < lat_len * dlat:
#                 raise CoordinateError("Variation in station latitude is "
#                                       "exceeding upper limit of {} m".format(
#                                       self.COORD_MAX_VAR['latitude']))
#             elif self.COORD_MAX_VAR['longitude'] < (lat_len *
#                                                     np.cos(np.deg2rad(lat)) * 
#                                                     dlon):
#                 raise CoordinateError("Variation in station longitude is "
#                                       "exceeding upper limit of {} m".format(
#                                       self.COORD_MAX_VAR['latitude']))
#             elif self.COORD_MAX_VAR['altitude'] < dalt:
#                 raise CoordinateError("Variation in station altitude is "
#                                       "exceeding upper limit of {} m".format(
#                                       self.COORD_MAX_VAR['latitude']))
# =============================================================================
        return vals
    
    def get_meta(self, force_single_value=True, quality_check=True,
                 add_none_vals=False):
        """Return meta-data as dictionary
        
        Parameters
        ----------
        force_single_value : bool
            if True, then each meta value that is list or array,is converted 
            to a single value. 
        quality_check : bool
            if True, and coordinate values are lists or arrays, then the 
            standarad deviation in the values is compared to the upper limits
            allowed in the local variation. The upper limits are specified
            in attr. ``COORD_MAX_VAR``. 
        add_none_vals : bool
            
        
        Returns
        -------
        dict
            dictionary containing the retrieved meta-data
            
        Raises
        ------
        AttributeError
            if one of the meta entries is invalid
        MetaDataError
            in case of consistencies in meta data between individual time-stamps
        """
        meta = {}
        meta.update(self.get_station_coords(force_single_value, quality_check))
        for key in self.STANDARD_META_KEYS:
            if key in self.STANDARD_COORD_KEYS: # this has been handled above
                continue
            if self[key] is None and not add_none_vals:
                logger.info('No metadata available for key {}'.format(key))
                continue
            
            val = self[key]
            if force_single_value and isinstance(val, (list, tuple, np.ndarray)):
                if quality_check and not all([x == val[0] for x in val]):
                    logger.debug("Performing quality check for meta data")
                    raise MetaDataError("Inconsistencies in meta parameter {} "
                                        "between different time-stamps".format(
                                        key))
                val = val[0]
            meta[key] = val
    
        return meta
    
    def _append_meta_item(self, key, val):
        """Add a metadata item"""
        if not key in self or self[key] == None:
            self[key] = val
        else:
            if isinstance(self[key], str):
                if not isinstance(val, str):
                    raise ValueError('Cannot merge meta item {} due to type '
                                     'mismatch'.format(key))
                vals = [x.strip() for x in self[key].split(';')]
                vals_in = [x.strip() for x in val.split(';')]
                
                for _val in vals_in:
                    if not _val in vals:
                        self[key] = self[key] + '; {}'.format(_val)
            else:
                if isinstance(val, (list, np.ndarray)):
                    raise ValueError('Cannot append metadata value that is '
                                     'already a list or numpy array due to '
                                     'potential ambiguities')
                if isinstance(self[key], list):
                    if not val in self[key]:
                        self[key].append(val)
                else:
                    if not self[key] == val:
                        self[key] = [self[key], val]
        return self
                        
    def merge_meta_same_station(self, other, coord_tol_km=None, 
                                check_coords=True, inplace=True,
                                **add_meta_keys):
        """Merge meta information from other object
        
        Note
        ----
        Coordinate attributes (latitude, longitude and altitude) are not 
        copied as they are required to be the same in both stations. The
        latter can be checked and ensured using input argument ``check_coords``
        
        Parameters
        ----------
        other : StationData
            other data object
        coord_tol_km : float
            maximum distance in km between coordinates of input StationData 
            object and self. Only relevant if :attr:`check_coords` is True. If
            None, then :attr:`_COORD_MAX_VAR` is used which is defined in the
            class header. 
        check_coords : bool
            if True, the coordinates are compared and checked if they are lying
            within a certain distance to each other (cf. :attr:`coord_tol_km`).
        inplace : bool
            if True, the metadata from the other station is added to the 
            metadata of this station, else, a new station is returned with the
            merged attributes.
        **add_meta_keys
            additional non-standard metadata keys that are supposed to be
            considered for merging.

                    
        """
        if not other.station_name == self.station_name:
            raise ValueError('Can only merged metadata from same station')
        
        if not inplace:
            from copy import deepcopy
            obj = deepcopy(self)
        else:
            obj = self
            
        if check_coords:
            if coord_tol_km is None:
                coord_tol_km = self._COORD_MAX_VAR
            try:
                self.same_coords(other, coord_tol_km)
            except MetaDataError: # 
                pass
                    
        keys = self.STANDARD_META_KEYS
        keys.extend(add_meta_keys)
        # remove station name from key list to be merged
        for key in keys:
            if key == 'station_name':
                continue
            if key in self.STANDARD_COORD_KEYS:
                if self[key] is None and other[key] is not None:
                    self[key] = other[key]
            
            elif key in other and other[key] is not None:
                obj._append_meta_item(key, other[key])
        
        return obj
        
    def merge_varinfo(self, other, var_name):
        """Merge variable specific meta information from other object
        
        Parameters
        ----------
        other : StationData
            other data object 
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
        """
        if not var_name in self.var_info:
            raise KeyError('No variable meta information available for {}'
                           .format(var_name))
            
        info_this = self.var_info[var_name]
        info_other = other.var_info[var_name]
        for key, val in info_other.items():
            if not key in info_this or info_this[key] == None:
                info_this[key] = val
            else:
                if isinstance(info_this[key], str):
                    if not isinstance(val, str):
                        raise ValueError('Cannot merge meta item {} due to type '
                                         'mismatch'.format(key))
                    vals = [x.strip() for x in info_this[key].split(';')]
                    vals_in = [x.strip() for x in val.split(';')]
                    
                    for _val in vals_in:
                        if not _val in vals:
                            info_this[key] = info_this[key] + ';{}'.format(_val)
                else:
                    if isinstance(val, (list, np.ndarray)):
                        if len(val) == 0:
                            continue
                        elif type(info_this[key]) == type(val):
                            if info_this[key] == val:
                                continue
                            info_this[key] = [info_this[key], val]
                        raise ValueError('Cannot append metadata value that is '
                                         'already a list or numpy array due to '
                                         'potential ambiguities')
                    if isinstance(info_this[key], list):
                        if not val in info_this[key]:
                            info_this[key].append(val)
                    else:
                        if not info_this[key] == val:
                            info_this[key] = [info_this[key], val]
        return self
# =============================================================================
#         for k, v in info_other.items():
#             if not k in info_this:
#                 info_this[k] = v
#             else:
#                 if not isinstance(info_this[k], list):
#                     info_this[k] = [info_this[k]]
#                     
#                 info_this[k].append(v)
# =============================================================================
        
    def check_if_3d(self, var_name):
        """Checks if altitude data is available in this object"""
        if 'altitude' in self: 
            val = self['altitude']
            if isnumeric(val): # is numerical value
                return False
            # unique altitude values
            uvals = np.unique(val)
            if len(uvals) == 1: # only one value in altitude array (NOT 3D)
                return False
            elif len(uvals[~np.isnan(uvals)]) == 1: # only 2 unique values in altitude array but one is NaN
                return False
            return True
        return False
        
    def _merge_vardata_3d(self, other, var_name):
        """Merge 3D variable data (for details see :func:`merge_vardata`)"""
        raise NotImplementedError
        s0 = self[var_name]
        s1 = other[var_name]
        info = other.var_info[var_name]
        removed = None
        if info['overlap']:
            raise NotImplementedError('Coming soon...')
        
        if len(s1) > 0: #there is data
            overlap = s0.index.intersection(s1.index)
            if len(overlap) > 0:
                removed = s1[overlap]
                s1 = s1.drop(index=overlap, inplace=True)
            
            #compute merged time series
            s0 = pd.concat([s0, s1], verify_integrity=True)
            
            # sort the concatenated series based on timestamps
            s0.sort_index(inplace=True)
            self.merge_varinfo(other, var_name)
        
        # assign merged time series (overwrites previous one)
        self[var_name] = s0
        
        if removed is not None:
            if var_name in self.overlap:
                self.overlap[var_name] = pd.concat([self.overlap[var_name], 
                                                   removed])
                self.overlap[var_name].sort_index(inplace=True)
            else:
                self.overlap[var_name] = removed
                
        return self

    def _ensure_same_var_ts_type_other(self, other, var_name):
        ts_type = self.get_var_ts_type(var_name)
        ts_type1 = other.get_var_ts_type(var_name)
        if ts_type != ts_type1:
            # make sure each variable in the object has explicitely ts_type
            # assigned (rather than global specification)

            self._update_var_timeinfo()
            other._update_var_timeinfo()

            from pyaerocom.helpers import get_lowest_resolution
            ts_type = get_lowest_resolution(ts_type, ts_type1)
        return ts_type

    def _update_var_timeinfo(self):

        for var, info in self.var_info.items():
            data = self[var]
            if not isinstance(data, pd.Series):
                try:
                    self[var] = pd.Series(data, self.dtime)
                except Exception as e:
                    raise Exception('Unexpected error: {}.\nPlease debug...'
                                    .format(repr(e)))
            if not 'ts_type' in info or info['ts_type'] is None:
                if not self.ts_type in const.GRID_IO.TS_TYPES:
                    raise ValueError('Cannot identify ts_type for var {} '
                                     'in {}'.format(var, self))
                info['ts_type'] = self.ts_type
        self.ts_type = None

    def _merge_vardata_2d(self, other, var_name):
        """Merge 2D variable data (for details see :func:`merge_vardata`)"""
        ts_type = self._ensure_same_var_ts_type_other(other, var_name)
        
        s0 = self.resample_timeseries(var_name, ts_type=ts_type,
                                      inplace=True).dropna()
        s1 = other.resample_timeseries(var_name, inplace=True,
                                       ts_type=ts_type).dropna()

        info = other.var_info[var_name]
        removed = None
        if 'overlap' in info and info['overlap']:
            raise NotImplementedError('Coming soon...')
        
        if len(s1) > 0: #there is data
            overlap = s0.index.intersection(s1.index)
            if len(overlap) > 0:
                removed = s1[overlap]
                s1 = s1.drop(index=overlap, inplace=True)
            #compute merged time series
            s0 = pd.concat([s0, s1], verify_integrity=True)
            
            # sort the concatenated series based on timestamps
            s0.sort_index(inplace=True)
            self.merge_varinfo(other, var_name)
        
        # assign merged time series (overwrites previous one)
        self[var_name] = s0
        self.dtime = s0.index.values
        
        if removed is not None:
            if var_name in self.overlap:
                self.overlap[var_name] = pd.concat([self.overlap[var_name], 
                                                   removed])
                self.overlap[var_name].sort_index(inplace=True)
            else:
                self.overlap[var_name] = removed
                
        return self
    
    def merge_vardata(self, other, var_name):
        """Merge variable data from other object into this object
        
        Note
        ----
        This merges also the information about this variable in the dict
        :attr:`var_info`. It is required, that variable meta-info is 
        specified in both StationData objects.
        
        Note
        ----
        This method removes NaN's from the existing time series in the data
        objects. In order to fill up the time-series with NaNs again after 
        merging, call :func:`insert_nans_timeseries`
        
        Parameters
        ----------
        other : StationData
            other data object 
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
            
        Returns
        -------
        StationData
            this object
        """
        if not var_name in self:
            raise VarNotAvailableError('StationData object does not contain '
                                       'data for variable {}'.format(var_name))
        elif not var_name in other:
            raise VarNotAvailableError('Input StationData object does not '
                                       'contain data for variable {}'.format(var_name))
        elif not var_name in self.var_info:
            raise MetaDataError('For merging of {} data, variable specific meta '
                                'data needs to be available in var_info dict '
                                .format(var_name))
        elif not var_name in other.var_info:
            raise MetaDataError('For merging of {} data, variable specific meta '
                                'data needs to be available in var_info dict '
                                .format(var_name))
        
        if self.get_unit(var_name) != other.get_unit(var_name):
            self.check_var_unit_aerocom(var_name)
            other.check_var_unit_aerocom(var_name)

        if self.check_if_3d(var_name):
            raise NotImplementedError('Coming soon...')
            #return self._merge_vardata_3d(other, var_name)
        else:
            return self._merge_vardata_2d(other, var_name)

    def merge_other(self, other, var_name, **add_meta_keys):
        """Merge other station data object
        
        Todo
        ----
        Should be independent of variable, i.e. it should be able to merge all
        data that is in the other object into this, even if this object does
        not contain that variable yet.
        
        Parameters
        ----------
        other : StationData
            other data object 
        var_name : str
            variable name for which info is to be merged (needs to be both
            available in this object and the provided other object)
            
        Returns
        -------
        StationData
            this object that has merged the other station
        """
        #self.merge_meta_same_station(other, **add_meta_keys)
        self.merge_vardata(other, var_name)
        self.merge_meta_same_station(other, **add_meta_keys)

        return self
        
    def get_data_columns(self):
        """List containing all data columns
        
        Iterates over all key / value pairs and finds all values that are 
        lists or numpy arrays that match the length of the time-stamp array 
        (attr. ``time``)
        
        Returns
        -------
        list
            list containing N arrays, where N is the total number of 
            datacolumns found. 
        """
        #self.check_dtime()
        #num = len(self.dtime)
        cols = {}
        for var_name in self.var_info:
            vals = self[var_name]
            if isinstance(vals, list):
                vals = np.asarray(vals)
            elif isinstance(vals, pd.Series):
                vals = vals.values
            elif isinstance(vals, VerticalProfile):
                raise NotImplementedError("This feature is not yet supported "
                                          "for data objects that contain also "
                                          "profile data")
            cols[var_name] = vals
        if not cols:
            raise AttributeError("No datacolumns could be found")
        return cols
    
    def check_dtime(self):
        """Checks if dtime attribute is array or list"""
        if not any([isinstance(self.dtime, x) for x in [list, np.ndarray]]):
            raise TypeError("dtime attribute is not iterable: {}".format(self.dtime))
        elif not len(self.dtime) > 0:
            raise AttributeError("No timestamps available")         
    
    def to_dataframe(self):
        """Convert this object to pandas dataframe
        
        Find all key/value pairs that contain observation data (i.e. values
        must be list or array and must have the same length as attribute 
        ``time``)
        
        """
        return pd.DataFrame(data=self.get_data_columns(), index=self.dtime)
    
    def get_var_ts_type(self, var_name, try_infer=True):
        """Get ts_type for a certain variable
        
        Note
        ----
        Converts to ts_type string if assigned ts_type is in pandas format
        
        Parameters
        ----------
        var_name : str
            data variable name for which the ts_type is supposed to be
            retrieved
        try_infer : bool
            if ts_type is not available, try inferring it from data
            
        Returns
        -------
        str
            the corresponding data time resolution
        
        Raises
        ------
        MetaDataError
            if no metadata is available for this variable (e.g. if ``var_name``
            cannot be found in :attr:`var_info`)
        """
        # make sure there exists a var_info dict for this variable
        if not var_name in self.var_info:
            self.var_info[var_name] = {}
        
        # use variable specific entry if available
        if 'ts_type' in self.var_info[var_name]:
            return TsType(self.var_info[var_name]['ts_type']).val
        elif isinstance(self.ts_type, str):
            # ensures validity and corrects for pandas strings
            ts_type = TsType(self.ts_type).val 
            self.var_info[var_name]['ts_type'] = ts_type
            return ts_type
        
        if try_infer:
            const.print_log.warning('Trying to infer ts_type in StationData {} '
                            'for variable {}'.format(self.station_name, var_name))
            from pyaerocom.helpers import infer_time_resolution
            try:
                s = self._to_ts_helper(var_name)
                ts_type = infer_time_resolution(s.index)
                self.var_info[var_name]['ts_type'] = ts_type
                return ts_type
            except:
                pass #Raise standard error
        raise MetaDataError('Could not access ts_type for {}'.format(var_name))
        
    def remove_outliers(self, var_name, low=None, high=None,
                        check_unit=True):
        """Remove outliers from one of the variable timeseries

        Parameters
        ----------
        var_name : str
            variable name
        low : float
            lower end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. minimum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        high : float
            upper end of valid range for input variable. If None, then the
            corresponding value from the default settings for this variable
            are used (cf. maximum attribute of `available variables
            <https://pyaerocom.met.no/config_files.html#variables>`__)
        check_unit : bool
            if True, the unit of the data is checked against AeroCom default
        """
        if any([x is None for x in (low, high)]):
            info = const.VARS[var_name]
            if check_unit:
                try:
                    self.check_unit(var_name)
                except DataUnitError:
                    self.convert_unit(var_name, to_unit=info.units)
            if low is None:
                low = info.minimum
                logger.info('Setting {} outlier lower lim: {:.2f}'
                            .format(var_name, low))
            if high is None:
                high = info.maximum
                logger.info('Setting {} outlier upper lim: {:.2f}'
                            .format(var_name, high))
            
            
        d = self[var_name]
        invalid_mask = np.logical_or(d<low, d>high)
        d[invalid_mask] = np.nan
        self[var_name] = d
        
    def interpolate_timeseries(self, var_name, freq, min_coverage_interp=0.3,
                               resample_how='mean', inplace=False):
        """Interpolate one variable timeseries to a certain frequency
        
        ToDo: complete docstring
        """
        raise NotImplementedError('Needs review...')
        new = self.to_timeseries(var_name, freq=freq, 
                                 resample_how='mean')
        coverage = 1 - new.isnull().sum() / len(new)
        if coverage < min_coverage_interp:
            from pyaerocom.exceptions import DataCoverageError
            raise DataCoverageError('{} data of station {} ({}) in '
                                    'time interval {} - {} contains '
                                    'too many invalid measurements '
                                    'for interpolation.'
                                    .format(var_name,
                                            self.station_name,
                                            self.data_id,
                                            self.dtime[0],
                                            self.dtime[-1]))
        new = new.interpolate().dropna()
        if inplace:
            ts_type = PANDAS_FREQ_TO_TS_TYPE[new.index.freqstr]
            self[var_name] = new
            
            self.var_info[var_name]['ts_type'] = ts_type        
            if len(self.var_info) > 1:     
                self.ts_type = None
            else:
                self.ts_type = ts_type
        return new
        
    def resample_timeseries(self, var_name, ts_type, how='mean',
                            apply_constraints=None, min_num_obs=None, 
                            inplace=False, **kwargs):
        """Resample one of the time-series in this object
        
        Parameters
        ----------
        var_name : str
            name of data variable
        ts_type : str
            new frequency string (can be pyaerocom ts_type or valid pandas
            frequency string)
        how : str
            how should the resampled data be averaged (e.g. mean, median)
        apply_constraints : bool, optional
            if True, then hierarchical resampling constraints are applied
            (for details see 
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        min_num_obs : dict or int, optional
            minimum number of observations required per period (when 
            downsampling). For details see 
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        inplace : bool
            if True, then the current data object stored in self, will be 
            overwritten with the resampled time-series
        **kwargs
            Additional keyword args passed to 
            :func:`pyaerocom.time_resampler.TimeResampler.resample`
            
        Returns
        -------
        pandas.Series or xarray.DataArray
            resampled time-series data
        """
        if not var_name in self:
            raise KeyError("Variable {} does not exist".format(var_name))
        
        to_ts_type = TsType(ts_type) # make sure to use AeroCom ts_type
        
        try: 
            from_ts_type = TsType(self.get_var_ts_type(var_name))
        except (MetaDataError, TemporalResolutionError):
            from_ts_type = None
            const.print_log.warning('Failed to access current temporal '
                                    'resolution of {} data in StationData {}. '
                                    'No resampling constraints will be applied'
                                    .format(var_name, self.station_name))
    
        data = self[var_name]
    
        if not isinstance(data, (pd.Series, xray.DataArray)):
            try:
                data = self.to_timeseries(var_name)
            except Exception as e:
                raise ValueError('{} data must be stored as pandas Series '
                                 'instance or as xarray.DataArray. Failed to '
                                 'convert to pandas Series.'
                                 'Error: {}'.format(repr(e)))
        resampler = TimeResampler(data)
        new = resampler.resample(to_ts_type=to_ts_type,
                                 from_ts_type=from_ts_type,
                                 how=how,
                                 apply_constraints=apply_constraints,
                                 min_num_obs=min_num_obs, 
                                 **kwargs)

# =============================================================================
#         if isinstance(data, pd.Series):
#             new = resample_timeseries(data, 
#                                       freq=to_ts_type.to_pandas(), 
#                                       how=how,
#                                       min_num_obs=min_num_obs)
#             
#         elif isinstance(data, xray.DataArray):
#             new = resample_time_dataarray(data, 
#                                           freq=to_ts_type.to_pandas(), 
#                                           how=how, 
#                                           min_num_obs=min_num_obs)
# =============================================================================
            
        if inplace:
            self[var_name] = new
            self.var_info[var_name]['ts_type'] = to_ts_type.val
            self.var_info[var_name].update(resampler.last_setup)
            # there is other variables that are not resampled
            if len(self.var_info) > 1 and self.ts_type is not None:     
                _tt = self.ts_type
                self.ts_type = None
                self.dtime = None
                for var, info in self.var_info.items():
                    if not var == var_name:
                        info['ts_type'] = _tt
            else: #no other variables, update global class attributes
                self.ts_type = to_ts_type.val
                self.dtime = new.index.values
        return new
    
    def remove_variable(self, var_name):
        """Remove variable data
        
        Parameters
        ----------
        var_name : str
            name of variable that is to be removed
            
        Returns
        -------
        StationData
            current instance of this object, with data removed
        
        Raises
        ------
        VarNotAvailableError
            if the input variable is not available in this object
        """
        if not self.has_var(var_name):
            raise VarNotAvailableError('No such variable in StationData: {}'
                                       .format(var_name))
        self.pop(var_name)
        if var_name in self.var_info:
            self.var_info.pop(var_name)
        return self
    
    def insert_nans_timeseries(self, var_name):
        """Fill up missing values with NaNs in an existing time series
        
        Note
        ----
        This method does a resample of the data onto a regular grid. Thus, if
        the input ``ts_type`` is different from the actual current ``ts_type`` 
        of the data, this method will not only insert NaNs but at the same.
        
        Parameters
        ---------
        var_name : str
            variable name
        inplace : bool
            if True, the actual data in this object will be overwritten with 
            the new data that contains NaNs
        
        Returns
        -------
        StationData
            the modified station data object
        
        """
        ts_type = self.get_var_ts_type(var_name)
        
        self.resample_timeseries(var_name, ts_type, inplace=True)
        
        return self

    
    def _to_ts_helper(self, var_name):
        """Convert data internally to pandas.Series if it is not stored as such
        
        Parameters
        ----------
        var_name : str
            variable name of data
        
        Returns
        -------
        pandas.Series
            data as timeseries
        """
        data = self[var_name]
        if isinstance(data, pd.Series):
            return data
        
        elif not data.ndim == 1:
            raise NotImplementedError('Multi-dimensional data columns '
                                      'cannot be converted to time-series')
        self.check_dtime()
        if not len(data) == len(self.dtime):
            raise ValueError("Mismatch between length of data array for "
                             "variable {} (length: {}) and time array  "
                             "(length: {}).".format(var_name, len(data), 
                               len(self.dtime)))    
        self[var_name] = s = pd.Series(data, index=self.dtime)
        return s
    
    def select_altitude(self, var_name, altitudes):
        
        data = self[var_name]
        
        if not isrange(altitudes):
            raise NotImplementedError('So far only a range (low, high) is '
                                      'supported for altitude extraction.')
            
        if isinstance(data, xray.DataArray):
            if not sorted(data.dims) == ['altitude', 'time']:
                raise NotImplementedError('Can only handle dataarrays that '
                                          'contain 2 dimensions altitude and '
                                          'time')
            if isrange(altitudes):
                if not isinstance(altitudes, slice):
                    altitudes = slice(altitudes[0], altitudes[1])
                return data.sel(altitude=altitudes)
            
            raise DataExtractionError('Cannot intepret input for '
                                          'altitude...')
        elif isinstance(data, pd.Series) and isinstance(data.index, pd.DatetimeIndex):
            if not len(data.index.unique()) == 1:
                raise DataDimensionError('Failed to interpret data dimensions')
            elif not 'altitude' in self:
                raise ValueError('Missing altitude information')
            elif not len(self.altitude) == len(data):
                raise DataDimensionError('Altitude data and {} data have '
                                         'different lengths'.format(var_name))
            
            alts = self['altitude']
            mask = np.logical_and(alts>=altitudes[0],
                                  alts<=altitudes[1])
            return data[mask]
            
                    
        
        raise DataExtractionError('Cannot extract altitudes: type of '
                                  '{} ({}) is not supported'
                                  .format(var_name, type(data)))
        
        
    def to_timeseries(self, var_name, freq=None, resample_how='mean', 
                      apply_constraints=None, min_num_obs=None, **kwargs):
        """Get pandas.Series object for one of the data columns
        
        Parameters
        ----------
        var_name : str
            name of variable (e.g. "od550aer")
        freq : str
            new temporal resolution (can be pandas freq. string, or pyaerocom
            ts_type)
        resample_how : str
            choose from mean or median (only relevant if input parameter freq 
            is provided, i.e. if resampling is applied)
        apply_constraints : bool, optional
            if True, then hierarchical resampling constraints are applied
            (for details see 
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        min_num_obs : dict or int, optional
            minimum number of observations required per period (when 
            downsampling). For details see 
            :func:`pyaerocom.time_resampler.TimeResampler.resample`)
        **kwargs
            optional keyword args passed to :func:`resample_timeseries`
            
        Returns
        -------
        Series
            time series object
        
        Raises 
        ------
        KeyError
            if variable key does not exist in this dictionary
        ValueError
            if length of data array does not equal the length of the time array
        """
        if not var_name in self:
            raise KeyError("Variable {} does not exist".format(var_name))
        data = self[var_name]
        alt_info = None
        # check if altitude subset is requested and process if applicable
        if 'altitude' in kwargs:
            alt_info =  kwargs.pop('altitude')
            data = self.select_altitude(var_name, alt_info)
        else:
            data = self[var_name]
        if 'ts_type' in kwargs:
            if freq is not None:
                raise ValueError('Both freq and ts_type are provided as input')
            freq = kwargs.pop('ts_type')
            
        if isinstance(data, xray.DataArray):
            if not 'time' in data.dims:
                raise NotImplementedError('Can only handle dataarrays that '
                                          'contain time dimension')
            if 'altitude' in data.dims: # collapsing of altitude dimension has not been done
                data = data.mean('altitude')
            if not data.ndim == 1:
                raise Exception('please debug')
            data = data.to_series()
        
        if not isinstance(data, pd.Series):
            data = self._to_ts_helper(var_name)

        if freq is not None:
            resampler = TimeResampler(data)
            try:
                from_ts_type = self.get_var_ts_type(var_name)
            except MetaDataError:
                from_ts_type = None
            data = resampler.resample(to_ts_type=freq, 
                                      from_ts_type=from_ts_type, 
                                      how=resample_how, 
                                      apply_constraints=apply_constraints,
                                      min_num_obs=min_num_obs,
                                      **kwargs)
# =============================================================================
#             data = resample_timeseries(data, freq, how=resample_how,
#                                        **kwargs)
# =============================================================================

        return data
    
    def plot_timeseries(self, var_name, freq=None, resample_how='mean', 
                        add_overlaps=False, legend=True, tit=None, 
                        **kwargs):
        """Plot timeseries for variable
        
        Note
        ----
        If you set input arg ``add_overlaps = True`` the overlapping timeseries
        data - if it exists - will be plotted on top of the actual timeseries
        using red colour and dashed line. As the overlapping data may be 
        identical with the actual data, you might want to increase the line 
        width of the actual timeseries using an additional input argumend 
        ``lw=4``, or similar.
        
        Parameters
        ----------
        var_name : str
            name of variable (e.g. "od550aer")
        freq : :obj:`str`, optional
            sampling resolution of data (can be pandas freq. string, or 
            pyaerocom ts_type).
        resample_how : :obj:`str`, optional
            choose from mean or median (only relevant if input parameter freq 
            is provided, i.e. if resampling is applied)
        add_overlaps : bool
            if True and if overlapping data exists for this variable, it will 
            be added to the plot.
        tit : :obj:`str`, optional
            title of plot, if None, default title is used
        **kwargs
            additional keyword args passed to matplotlib ``plot`` method
            
        Returns
        -------
        axes
            matplotlib.axes instance of plot
        
        Raises 
        ------
        KeyError
            if variable key does not exist in this dictionary
        ValueError
            if length of data array does not equal the length of the time array
        """
        import matplotlib.pyplot as plt
        if 'label' in kwargs:
            lbl = kwargs.pop('label')
        else:
            lbl = var_name
            if freq is not None:
                lbl += ' ({} {})'.format(freq, resample_how)
            else:
                try:
                    ts_type = self.get_var_ts_type(var_name)
                    lbl += ' ({})'.format(ts_type)
                except:
                    pass
        if not 'ax' in kwargs:
            if 'figsize' in kwargs:
                fs = kwargs.pop('figsize')
            else:
                fs = (16, 8)
            _, ax = plt.subplots(1, 1, figsize=fs)
        else: 
            ax = kwargs.pop('ax')
            # keep existing title if it exists
            _tit = ax.get_title()
            if not _tit == '':
                tit = _tit
        
        if tit is None:
            try:
                tit = self.get_meta(force_single_value=True, 
                                      quality_check=False)['station_name']    
            except:
                tit = 'Failed to retrieve station_name'
        s = self.to_timeseries(var_name, freq, resample_how)
        
        ax.plot(s, label=lbl, **kwargs)
        if add_overlaps and var_name in self.overlap:
            so = self.overlap[var_name]
            try:
                from pyaerocom.helpers import resample_timeseries
                so = resample_timeseries(so, freq, how=resample_how)
            except:
                pass
            if var_name in self.overlap:
                ax.plot(so, '--', lw=1, c='r', 
                        label='{} (overlap)'.format(var_name)) 
            else: 
                tit += ' (No overlapping data found)'
                
        ylabel = var_name
        try:
            if 'units' in self.var_info[var_name]:
                u = self.var_info[var_name]['units']
                if u is not None and not u in [1, '1']:
                    ylabel += ' [{}]'.format(u)
        except:
            logger.warning('Failed to access unit information for variable {}'
                           .format(var_name))
        ax.set_ylabel(ylabel)
        ax.set_title(tit)
        if legend:
            ax.legend()
        return ax

    def __str__(self):
        """String representation"""
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        arrays = ''
        series = ''
    
        for k, v in self.items():
            if k[0] == '_':
                continue
            if isinstance(v, dict):
                s += "\n{} ({}):".format(k, type(v).__name__)
                if v:
                    s = dict_to_str(v, s, indent=2)
                else:
                    s += ' <empty_dict>'
            elif isinstance(v, list):
                s += list_to_shortstr(v, name=k)
            elif isinstance(v, np.ndarray):
                if v.ndim==1:
                    arrays += list_to_shortstr(v, name=k)
                else:
                    arrays += "\n{} (ndarray, shape {})".format(k, v.shape)
                    arrays += "\n{}".format(v)
            elif isinstance(v, pd.Series):
                series += "\n{} (Series, {} items)".format(k, len(v))
            else:
                if v == '':
                    v = '<empty_str>'
                s += "\n{}: {}".format(k,v)
        if arrays:
            s += '\n\nData arrays\n.................'
            s += arrays
        if series:
            s += '\nPandas Series\n.................'
            s += series
    
        return s
    
if __name__=="__main__":
    import pyaerocom as pya
    import matplotlib.pyplot as plt
    
    plt.close('all')
    data0 = pya.io.ReadUngridded().read('AeronetSunV3Lev2.daily', 'od550aer')
    
    sv = data0.to_station_data('Solar*')
    sv.plot_timeseries('od550aer', 'yearly')
    
    data = pya.io.ReadUngridded().read('EBASMC', 'concoc')
    
    birkenes = data.to_station_data('Birken*')
    
    ax = birkenes.plot_timeseries('concoc')
    ax = birkenes.plot_timeseries('concoc', 'yearly', ax=ax)
    


#!/usr/bin/env python3

################################################################
# plotmodelmaps.py
#
# model data map plotting
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171130 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

#Copyright (C) 2017 met.no
#Contact information:
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA


#import pdb
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap

import iris.coord_categorisation
import iris.plot as iplt

import cartopy.crs as ccrs
from cartopy.mpl.ticker import LongitudeFormatter, LatitudeFormatter

import cf_units as unit
import os

from pyaerocom.griddeddata import GriddedData

def plotmaps(data, verbose = False, filter ="WORLD", var="od550aer",
             plotdir="./"):
    """plot aerocom standard maps

    Will plot every supplied time step"""

    if not isinstance(data, GriddedData):
        raise TypeError("Need pyaerocom.GriddedData as input")
    #define color bar;
    #will be moved somewhere else and variable specific at some point
    colorbar_levels = [0., 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8,
                       0.9, 1.0]

    colorbar_ticks = [0., 0.02, 0.04, 0.06, 0.08, 0.1, 0.3, 0.5, 0.7, 0.9]
    aod_data = {
        'red': ((0., 0, 0), (0.07, 0, 0), (0.1, 0.678, 0.678,), (0.26, 1, 1), (0.85, 1, 1), (1, 0.545, 0.545)),
        'green': ((0., 0, 0), (0.07, 1, 1), (0.24, 1, 1), (0.91, 0, 0), (1, 0, 0)),
        'blue': ((0., 1, 1), (0.07, 1, 1), (0.1, 0.133, 0.133), (0.25, 0, 0), (1, 0, 0))}

    colormap = LinearSegmentedColormap('aod_jet', aod_data)
    plt.register_cmap(cmap = colormap)
    TIME_VAR_NAME = 'time'

    PlotDailyFlag = False
    PlotMonthlyFlag = True

    for model in data:
        #assume that we have single cube for the moment
        #iterate over the time dimension
        cube = data[model].data
        #model = 'AATSR_ORAC_v4.01'

        cube.coord('latitude').guess_bounds()
        cube.coord('longitude').guess_bounds()

        #plot daily data
        if PlotDailyFlag:
            for time_sub_cube in cube.slices_over(TIME_VAR_NAME):
                pre_grid_areas = iris.analysis.cartography.area_weights(time_sub_cube)
                #pdb.set_trace()
                #print(time_sub_cube)

                # Perform the area-weighted mean for each of the datasets using the
                # computed grid-box areas.
                weighted_mean = time_sub_cube.collapsed(['latitude', 'longitude'],
                                                        iris.analysis.MEAN,
                                                        weights = pre_grid_areas)

                time = (unit.num2date(time_sub_cube.coord(TIME_VAR_NAME).points,
                                      time_sub_cube.coord(TIME_VAR_NAME).units.name,
                                      time_sub_cube.coord(TIME_VAR_NAME).units.calendar))

                date = str(time[0]).split(' ')[0].replace('-','')
                Year = date[0:4]
                TsStr = 'm'+date
                PlotType = 'MAP'

                title = " ".join([var, date, 'mean: {:6.3f}'.format(float(weighted_mean.data))])
                #ABS550_AER_an2012_mALLYEAR_WORLD_ZONALOBS_AERONETSky.ps.png
                #OD550_AER_an2017_d20171125_WORLD_MAP.ps.png
                plotfilename = os.path.join(plotdir, '_'.join([var, "an" + Year, TsStr, filter, PlotType]) + ".png")
                if verbose:
                    print(plotfilename)
                plot = iplt.pcolormesh(time_sub_cube[:, :], cmap = colormap, vmin=0., vmax=max(colorbar_levels))
                # pdb.set_trace()
                plot.axes.set_aspect(1.8)
                LatsToPlot = time_sub_cube.coord(axis='X').points
                LonsToPlot = time_sub_cube.coord(axis='Y').points
                axis = plt.axis([LatsToPlot.min(), LatsToPlot.max(), LonsToPlot.min(), LonsToPlot.max()])
                ax = plot.axes
                ax.annotate('source: AEROCOM', xy=(0.88, 0.04), xycoords='figure fraction',
                            horizontalalignment='right', fontsize=10, bbox=dict(boxstyle='square', facecolor='none',
                                                                                edgecolor='black'))
                ax.annotate(model, xy=(-174., -83.), xycoords='data', horizontalalignment='left', fontsize=13,
                            color='black', bbox=dict(boxstyle='square', facecolor='white', edgecolor='none', alpha=0.7))

                # plt.ylabel(_title(plot_defn.coords[0], with_units=True))
                # plot_defn = iplt._get_plot_defn(cube, mode, ndims)

                plt.colorbar(spacing='uniform', ticks=colorbar_ticks, 
                             boundaries=colorbar_levels, extend='max')

                ax.coastlines()
                ax.set_xticks([-180., -120., -60., 0., 60, 120, 180], crs=ccrs.PlateCarree())
                ax.set_yticks([-90., -60, -30, 0., 30, 60, 90], crs=ccrs.PlateCarree())
                lon_formatter = LongitudeFormatter(number_format='.1f', degree_symbol='')
                lat_formatter = LatitudeFormatter(number_format='.1f', degree_symbol='')
                ax.xaxis.set_major_formatter(lon_formatter)
                ax.yaxis.set_major_formatter(lat_formatter)

                plt.xlabel = 'longitude'
                plt.ylabel = 'latitude'

                plt.title(title)
                plt.savefig(plotfilename, dpi=300)
                plt.close()
                #pdb.set_trace()

        elif PlotMonthlyFlag:
            #calculate monthly data

            iris.coord_categorisation.add_month(cube, TIME_VAR_NAME, name='month_number')
            iris.coord_categorisation.add_year(cube, TIME_VAR_NAME, name='month_year')
            cube_monthly = cube.aggregated_by(['month_number', 'month_year'], iris.analysis.MEAN)

            for time_sub_cube in cube_monthly.slices_over(TIME_VAR_NAME):
                pre_grid_areas = iris.analysis.cartography.area_weights(time_sub_cube)
                # pdb.set_trace()
                # print(time_sub_cube)

                # Perform the area-weighted mean for each of the datasets using the
                # computed grid-box areas.
                weighted_mean = time_sub_cube.collapsed(['latitude', 'longitude'],
                                                        iris.analysis.MEAN,
                                                        weights=pre_grid_areas)

                time = (unit.num2date(time_sub_cube.coord(TIME_VAR_NAME).points,
                                      time_sub_cube.coord(TIME_VAR_NAME).units.name,
                                      time_sub_cube.coord(TIME_VAR_NAME).units.calendar))

                date = str(time[0]).split(' ')[0].replace('-', ' ')[0:7]
                Year = date[0:4]
                TsStr = 'm' + date[-2:]
                PlotType = 'MAP'

                title = " ".join([var, date, 'mean: {:6.3f}'.format(float(weighted_mean.data))])
                # ABS550_AER_an2012_mALLYEAR_WORLD_ZONALOBS_AERONETSky.ps.png
                # OD550_AER_an2017_d20171125_WORLD_MAP.ps.png
                plotfilename = os.path.join(plotdir, '_'.join([var, "an" + Year, TsStr, filter, PlotType]) + ".png")
                if verbose:
                    print(plotfilename)
                LatsToPlot = time_sub_cube.coord(axis='X').points
                LonsToPlot = time_sub_cube.coord(axis='Y').points
                xticks = [-180., -120., -60., 0., 60, 120, 180]
                yticks = [-90., -60, -30, 0., 30, 60, 90]
                plot_ts_map(time_sub_cube, title, plotfilename, LatsToPlot, LonsToPlot, colorbar_ticks,
                            colorbar_levels, colormap, model, xticks, yticks)

def annotate_aerocom(ax):
    """annotate with source: aerocom
    :rtype: None
    """

    ax.annotate('source: AEROCOM', xy=(0.88, 0.04), xycoords='figure fraction',
                horizontalalignment='right', fontsize=10, bbox=dict(boxstyle='square', facecolor='none',
                                                                    edgecolor='black'))

def plot_ts_map(data, title, plotfilename, lats_to_plot, lons_to_plot, colorbar_ticks, colorbar_levels,
                colormap, model_name,xticks, yticks,
                VerboseFlag = False):

    """function to plot one time step of a cube for aerocom"""

    plot = iplt.pcolormesh(data, cmap=colormap, vmin=0., vmax=max(colorbar_levels))
    plot.axes.set_aspect(1.8)
    plot.axes.axis([lats_to_plot.min(), lats_to_plot.max(), lons_to_plot.min(), lons_to_plot.max()])
    # axis = plt.axis([LatsToPlot.min(), LatsToPlot.max(), LonsToPlot.min(), LonsToPlot.max()])
    ax = plot.axes
    annotate_aerocom(ax)
    ax.annotate(model_name, xy=(-174., -83.), xycoords='data', horizontalalignment='left', fontsize=13,
                color='black', bbox=dict(boxstyle='square', facecolor='white', edgecolor='none', alpha=0.7))

    # plt.ylabel(_title(plot_defn.coords[0], with_units=True))
    # plot_defn = iplt._get_plot_defn(cube, mode, ndims)

    plt.colorbar(spacing='uniform', ticks=colorbar_ticks, boundaries=colorbar_levels, extend='max')

    ax.coastlines()
    ax.set_xticks(xticks, crs=ccrs.PlateCarree())
    ax.set_yticks(yticks, crs=ccrs.PlateCarree())
    # ax.set_xticks([-180., -120., -60., 0., 60, 120, 180], crs=ccrs.PlateCarree())
    # ax.set_yticks([-90., -60, -30, 0., 30, 60, 90], crs=ccrs.PlateCarree())
    lon_formatter = LongitudeFormatter(number_format='.1f', degree_symbol='')
    lat_formatter = LatitudeFormatter(number_format='.1f', degree_symbol='')
    ax.xaxis.set_major_formatter(lon_formatter)
    ax.yaxis.set_major_formatter(lat_formatter)

    plot.axes.set_title(title)
    plot.axes.set_xlabel('longitude')
    plot.axes.set_ylabel('latitude')
    plt.savefig(plotfilename, dpi=300)
    #pdb.set_trace()
    plt.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Nov 16 10:31:09 2018

@author: jonasg
"""

try:
    import ipywidgets
    ipw_avail = True
except:
    ipw_avail = False
    
if ipw_avail:
    from .ipywidgets import ProgressBarLabelled


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
import numpy as np
import numpy.testing as npt
from pandas import Timestamp
from pyaerocom import ColocatedData

def test_meta_access_filename():
    name = 'absc550aer_REF-EBAS-Lev3_MOD-CAM5-ATRAS_20100101_20101231_daily_WORLD-noMOUNTAINS.nc'
    
    meta = {'var_name': 'absc550aer', 
            'ts_type': 'daily', 
            'filter_name': 'WORLD-noMOUNTAINS', 
            'start': Timestamp('2010-01-01 00:00:00'), 
            'stop': Timestamp('2010-12-31 00:00:00'), 
            'data_source': ['EBAS-Lev3', 'CAM5-ATRAS']}
    for k, v in ColocatedData.get_meta_from_filename(name).items():
        assert meta[k] == v
    
    
if __name__=="__main__":
    
    test_meta_access_filename()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Config file for AeroCom PhaseIII test project
"""


import iris
import numpy as np
from pyaerocom import print_log
from pyaerocom._lowlevel_helpers import merge_dicts
from pyaerocom.helpers import copy_coords_cube

CUBE_MATHS = {'add'         :   iris.analysis.maths.add,
              'subtract'    :   iris.analysis.maths.subtract,
              'divide'      :   iris.analysis.maths.divide,
              'multiply'    :   iris.analysis.maths.multiply}

def _apply_operator_cubes(cube1, cube2, operator_name,
                          allow_coord_merge=True):
    if not operator_name in CUBE_MATHS:
        raise NotImplementedError('No such arithmetic cube operator '
                                  'implemented: {}'.format(operator_name))
    fun = CUBE_MATHS[operator_name]
    try:
        return fun(cube1, cube2)
    except ValueError as e:
        print_log.warning('Could not {} cubes straight out of the box. Trying '
                          'to correct for dimension definition errors'
                          .format(operator_name))
        if 'differing coordinates (time)' in repr(e):
            iris.util.unify_time_units([cube1, cube2])    
        
        elif allow_coord_merge:
            copy_coords_cube(to_cube=cube2, 
                             from_cube=cube1, 
                             inplace=True)
    return fun(cube1, cube2)

def _check_input_iscube(*data_objs):
    from pyaerocom.griddeddata import GriddedData
    checked = []
    for obj in data_objs:
        if isinstance(obj, GriddedData):
            checked.append(obj.cube)
        elif isinstance(obj, iris.cube.Cube):
            checked.append(obj)
        else:
            raise ValueError('Invalid input: require GriddedData or Cube, got '
                             '{}'.format(type(obj)))
    return checked

def _check_same_units(cube1, cube2):
    if cube1.units == cube2.units:
        return (cube1, cube2)
    from pyaerocom import const
    var1 = cube1.var_name
    u1 = cube1.units
    
    var2 = cube2.var_name
    u2 = cube2.units

    if var1 in const.VARS and u1 == const.VARS[var1]['units'] :
        cube2.convert_units(u1)
        return (cube1, cube2)
    elif var2 in const.VARS and u2 == const.VARS[var2]['units'] :
        cube1.convert_units(u2)
        return (cube1, cube2)
    
    try:
        cube2.convert_units(u1)
        return (cube1, cube2)
    except:
        from pyaerocom.exceptions import UnitConversionError
        raise UnitConversionError('Failed to harmonise units')

def merge_meta_cubes(cube1, cube2):
    try:
        return merge_dicts(cube1.attributes,
                           cube2.attributes)
    except Exception as e:
        print_log.warning('Failed to merge Cube metadata. Error: {}'
                          .format(repr(e)))
        return {'NOTE'  : 'MERGE_FAILED',
                'meta1' : cube1.attributes, 
                'meta2' : cube2.attributes}

def apply_rh_thresh_cubes(cube, rh_cube, rh_max=None):
    """Method that applies a low RH filter to input cube"""
    cube, rh_cube =  _check_input_iscube(cube, rh_cube)
    if rh_max is None:
        from pyaerocom import const
        rh_max = const.VARS[cube.var_name]['dry_rh_max']
    if not cube.shape == rh_cube.shape:
        raise ValueError
    mask = rh_cube.data > rh_max
    cube_out = cube.copy()
    cube_out.data[mask] = np.nan
    cube_out.attributes.update(cube.attributes)
    cube_out.attributes['rh_max'] = rh_max
    return cube_out
    
        
def add_cubes(cube1, cube2):
    """Method to add cubes from 2 gridded data objects
    """
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)
    
    cube_out = _apply_operator_cubes(cube1, cube2, 'add', 
                                     allow_coord_merge=True)
    
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def subtract_cubes(cube1, cube2):
    """Method to subtract 1 cube from another"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube1, cube2 = _check_same_units(cube1, cube2)    
    
    cube_out = _apply_operator_cubes(cube1, cube2, 'subtract', 
                                     allow_coord_merge=True)
    
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out
   
def multiply_cubes(cube1, cube2):
    """Method to multiply 2 cubes"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, 'multiply', 
                                     allow_coord_merge=True)
    
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out

def divide_cubes(cube1, cube2):
    """Method to divide 2 cubes with each other"""
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    cube_out = _apply_operator_cubes(cube1, cube2, 'divide', 
                                     allow_coord_merge=True)
    
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out
    
def compute_angstrom_coeff_cubes(cube1, cube2, lambda1=None, lambda2=None):
    """Compute Angstrom coefficient cube based on 2 optical densitiy cubes
    
    Parameters
    ----------
    cube1 : iris.cube.Cube
        AOD at wavelength 1
    cube2 : iris.cube.Cube
        AOD at wavelength 2
    lambda1 : float
        wavelength 1
    lambda 2 : float
        wavelength 2
        
    Returns
    -------
    Cube
        Cube containing Angstrom exponent(s)
    """
    from pyaerocom.variable import VarNameInfo
    from cf_units import Unit
    cube1, cube2 = _check_input_iscube(cube1, cube2)
    if lambda1 is None:
        lambda1 = VarNameInfo(cube1.var_name).wavelength_nm
    if lambda2 is None:
        lambda2 = VarNameInfo(cube2.var_name).wavelength_nm
    
    ratio = _apply_operator_cubes(cube1, cube2, 'divide')
    logr = iris.analysis.maths.log(ratio)
    
    wvl_r = np.log(lambda1 / lambda2)
    cube_out = -1*iris.analysis.maths.divide(logr, wvl_r)
    cube_out.units = Unit('1')
    cube_out.attributes.update(merge_meta_cubes(cube1, cube2))
    return cube_out


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains scatter plot routines for Aerocom data.
"""
from matplotlib.ticker import ScalarFormatter
import os
import matplotlib.pyplot as plt
import numpy as np
from pyaerocom import const
from pyaerocom.helpers import start_stop_str
from pyaerocom.mathutils import calc_statistics, exponent

def plot_scatter(x_vals, y_vals, **kwargs):
    """Scatter plot
    
    Currently a wrapper for high-level method plot_scatter_aerocom (same module,
    see there for details)
    """
    return plot_scatter_aerocom(x_vals, y_vals, **kwargs)

def plot_scatter_aerocom(x_vals, y_vals, var_name=None, var_name_ref=None, 
                         x_name=None, y_name=None, start=None, stop=None, 
                         ts_type=None, unit=None, stations_ok=None, 
                         filter_name=None, lowlim_stats=None, 
                         highlim_stats=None, loglog=None, savefig=False, 
                         save_dir=None, save_name=None, ax=None, figsize=None, 
                         fontsize_base=10, marker='+', color='k', alpha=0.5,
                         **kwargs):
    """Method that performs a scatter plot of data in AEROCOM format
    
    Parameters
    ----------
    y_vals : ndarray
        1D array (or list) of model data points (y-axis)
    x_vals : ndarray
        1D array (or list) of observation data points (x-axis)
    var_name : :obj:`str`, optional
        name of variable that is plotted
    var_name_ref : :obj:`str`, optional
        name of variable of reference data
    x_name : :obj:`str`, optional
        Name of observation network
    y_name : :obj:`str`, optional
        Name / ID of model
    start : :obj:`str` or :obj`datetime` or similar
        start time of data
    stop : :obj:`str` or :obj`datetime` or similar
        stop time of data
    
    Returns 
    -------
    axes
        instance of :class:`matplotlib.axes`
    """

    if isinstance(y_vals, list):
        y_vals = np.asarray(y_vals)
    if isinstance(x_vals, list):
        x_vals = np.asarray(x_vals)
    try:
        VARS = const.VARS[var_name]
    except:
        VARS = const.VARS.DEFAULT
    if loglog is None:
        loglog = VARS.scat_loglog
        
    xlim = VARS['scat_xlim']
    ylim = VARS['scat_ylim']
    
    if xlim is None or ylim is None:
        low  =  np.min([np.nanmin(x_vals), np.nanmin(y_vals)])
        high =  np.max([np.nanmax(x_vals), np.nanmax(y_vals)])
        
        xlim = [low, high]
        ylim = [low, high]

    if ax is None:
        if figsize is None:
            figsize = (10,8)
        fig, ax = plt.subplots(figsize=figsize)
    if var_name is None:
        var_name = 'n/d'
    
    statistics = calc_statistics(y_vals, x_vals,
                                 lowlim_stats, highlim_stats)

    if loglog:
        ax.loglog(x_vals, y_vals, ls='none', color=color, marker=marker, 
                  alpha=alpha, **kwargs)
    else:
        ax.plot(x_vals, y_vals, ls='none', color=color, marker=marker, 
                  alpha=alpha, **kwargs)
    
    try:
        title = start_stop_str(start, stop, ts_type)
        if ts_type is not None:
            title += ' ({})'.format(ts_type)
    except:
        title = ''
    
    if not loglog:
        xlim[0] = 0
        ylim[0] = 0
    elif any(x[0] < 0 for x in [xlim, ylim]):
        
        low = 10**(float(exponent(abs(np.nanmin(y_vals))) - 1))
        xlim[0] = low
        ylim[0] = low
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    xlbl = '{}'.format(x_name)
    if var_name_ref is not None:
        xlbl += ' ({})'.format(var_name_ref)
    ax.set_xlabel(xlbl, fontsize=fontsize_base+4)
    ax.set_ylabel('{}'.format(y_name), fontsize=fontsize_base+4)
    
    ax.set_title(title, fontsize=fontsize_base+4)
    
    ax.xaxis.set_major_formatter(ScalarFormatter())
    ax.yaxis.set_major_formatter(ScalarFormatter())
    
    ax.tick_params(labelsize=fontsize_base)
    
    ax.plot(xlim, ylim, '-', color='grey')
    
    xypos =   {'var_info'       :   (0.01, .95),
               'refdata_mean'   :   (0.01, 0.90),
               'data_mean'      :   (0.01, 0.86),
               'nmb'            :   (0.01, 0.82),
               'mnmb'           :   (0.35, 0.82),
               'R'              :   (0.01, 0.78),
               'rms'            :   (0.35, 0.78),
               'R_kendall'      :   (0.01, 0.74),
               'fge'            :   (0.35, 0.74),
               'ts_type'        :   (0.8, 0.1),
               'filter_name'    :   (0.8, 0.06)}
    
    var_str = var_name# + VARS.unit_str
    _ndig = abs(exponent(statistics['refdata_mean']) - 2)
    if unit is None:
        unit = 'N/D'
    if not str(unit) in ['1', 'no_unit']:
        var_str += ' [{}]'.format(unit)

    ax.annotate("{} #: {} # st: {}".format(var_str, 
                        statistics['num_valid'], stations_ok),
                        xy=xypos['var_info'], xycoords='axes fraction', 
                        fontsize=fontsize_base+4, color='red')

    ax.annotate('Mean (x-data): {:.{}f}; Rng: [{:.{}f}, {:.{}f}]'
                .format(statistics['refdata_mean'], _ndig,
                        np.nanmin(x_vals),_ndig,
                        np.nanmax(x_vals), _ndig),
                xy=xypos['refdata_mean'], xycoords='axes fraction', 
                fontsize=fontsize_base, 
                color='red')
    
    ax.annotate('Mean (y-data): {:.{}f}; Rng: [{:.{}f}, {:.{}f}]'
                .format(statistics['data_mean'], _ndig,
                        np.nanmin(y_vals),_ndig,
                        np.nanmax(y_vals), _ndig),
                xy=xypos['data_mean'], xycoords='axes fraction', 
                fontsize=fontsize_base, 
                color='red')
    
    ax.annotate('NMB: {:.1f}%'.format(statistics['nmb']*100),
                        xy=xypos['nmb'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')
    
    ax.annotate('MNMB: {:.1f}%'.format(statistics['mnmb']*100),
                        xy=xypos['mnmb'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')
    
    ax.annotate('R (Pearson): {:.3f}'.format(statistics['R']),
                        xy=xypos['R'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')

    ax.annotate('RMS: {:.3f}'.format(statistics['rms']),
                        xy=xypos['rms'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')
    
    ax.annotate('R (Kendall): {:.3f}'.format(statistics['R_kendall']),
                        xy=xypos['R_kendall'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')
    
    
    ax.annotate('FGE: {:.1f}'.format(statistics['fge']),
                        xy=xypos['fge'], xycoords='axes fraction', 
                        fontsize=fontsize_base, color='red')
    # right lower part
    ax.annotate('{}'.format(ts_type),
                        xy=xypos['ts_type'], xycoords='axes fraction', 
                        ha='center', 
                        fontsize=fontsize_base, color='black')
    ax.annotate('{}'.format(filter_name),
                        xy=xypos['filter_name'], xycoords='axes fraction', 
                        ha='center', 
                        fontsize=fontsize_base, color='black')
    
    ax.set_aspect('equal')
    
    if savefig:
        if any([x is None for x in (save_dir, save_name)]):
            raise IOError
            
        fig.savefig(os.path.join(save_dir, save_name))
    return ax
    




################################################################
# plot/__init__.py
#
# init for map plotting
#
# this file is part of the aerocom_pt package
#
#################################################################
# Created 20171130 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

#Copyright (C) 2017 met.no
#Contact information:
#Norwegian Meteorological Institute
#Box 43 Blindern
#0313 OSLO
#NORWAY
#E-mail: jan.griesfeller@met.no
#This program is free software; you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation; either version 3 of the License, or
#(at your option) any later version.
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#GNU General Public License for more details.
#You should have received a copy of the GNU General Public License
#along with this program; if not, write to the Free Software
#Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#MA 02110-1301, USA


# Note that these are temporary imports, namespaces will be redefined in the
# near future
from .plotmaps import plotmaps, annotate_aerocom, plot_ts_map
from .mapping import plot_griddeddata_on_map
from .plotscatter_v0 import plotscatter
from .plotseries import plot_series_year
from .heatmaps import df_to_heatmap
from .plotsitelocation import plotsitelocation



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
High level I/O utility methods for pyaerocom
"""

from pyaerocom.io.aerocom_browser import AerocomBrowser
from pyaerocom.io.readgridded import ReadGridded
from pyaerocom.io.readungridded import ReadUngridded
from pyaerocom import const, change_verbosity

def browse_database(model_or_obs, verbose=False):
    """Browse Aerocom database using model or obs ID (or wildcard)
    
    Searches database for matches and prints information about all matches 
    found (e.g. available variables, years, etc.)
    
    Parameters
    ----------
    model_or_obs : str
        model or obs ID or search pattern 
    verbose : bool
        if True, verbosity level will be set to debug, else to critical
        
    Example
    -------
    >>> import pyaerocom as pya
    >>> pya.io.browse_database('AATSR*ORAC*v4*')
    Pyaerocom ReadGridded
    ---------------------
    Model ID: AATSR_ORAC_v4.01
    Data directory: /lustre/storeA/project/aerocom/aerocom-users-database/CCI-Aerosol/CCI_AEROSOL_Phase2/AATSR_ORAC_v4.01/renamed
    Available variables: ['abs550aer', 'ang4487aer', 'clt', 'landseamask', 'od550aer', 'od550dust', 'od550gt1aer', 'od550lt1aer', 'pixelcount']
    Available years: [2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010, 2011, 2012]
    Available time resolutions ['daily']
    """
    if not verbose:
        change_verbosity('critical')
    else:
        change_verbosity('debug')
    browser = AerocomBrowser()
    matches = browser.find_matches(model_or_obs)
    if len(matches) == 0:
        print('No match could be found for {}'.format(model_or_obs))
        return
    elif len(matches) > 20:
        print('Found more than 20 matches for input pattern {}:\n\n'
              'Matches: {}\n\n'
              'To receive more detailed information, please specify search ID '
              'more accurately'.format(model_or_obs, matches))
        return
    for match in matches:
        try:
            if match in const.OBS_IDS:
                reader = ReadUngridded(match)
            else:
                reader = ReadGridded(match)
            print(reader)
        except Exception as e:
            print('Reading failed for {}. Error: {}'.format(match,
                  repr(e)))
    

if __name__=='__main__':
    
    obs_id = 'AATSR*'
    
    browse_database('AATSR_SU*')
    
    browse_database('AATSR*ORAC*v4*')
    
    browse_database(obs_id)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr 12 14:45:43 2018

@author: jonasg
"""
import numpy as np
import pytest
from pyaerocom.test.settings import DATA_ACCESS
#import numpy.testing as npt
from pyaerocom import helpers

def test_get_standarad_name():
    assert (helpers.get_standard_name('od550aer') == 
            'atmosphere_optical_thickness_due_to_ambient_aerosol_particles')

def test_get_standard_unit():
    assert helpers.get_standard_unit('ec550aer') == '1/Mm'

def test_get_lowest_resolution():
    assert helpers.get_lowest_resolution('3hourly', 
                                         'hourly', 
                                         'monthly', 
                                         'yearly') == 'yearly'
def test_isnumeric():
    assert helpers.isnumeric(3)
    assert helpers.isnumeric(3.3455)
    assert helpers.isnumeric(np.complex(1,2))
    
def test_isrange():
    assert helpers.isrange((0, 1))
    assert helpers.isrange([10, 20])
        
def test_merge_station_data():
    pass

def test__get_pandas_freq_and_loffset():
    val = helpers._get_pandas_freq_and_loffset('monthly')
    assert val == ('MS', '14D')


def _make_timeseries_synthetic():
    from pandas import Series, DatetimeIndex
    idx = DatetimeIndex(freq='D', start='2000', periods=90)
    vals = np.arange(len(idx))
    
    return  Series(vals, idx)

@pytest.fixture(scope='module')
def timeseries_synthetic():
    return _make_timeseries_synthetic()\

def test_resample_timeseries(timeseries_synthetic):
    s1 = helpers.resample_timeseries(timeseries_synthetic, 'monthly')
    
    assert len(s1) == 3
    for time in s1.index:
        assert time.year == 2000, time.year
        assert time.day == 15, time.day
    
def test_same_meta_dict():
    d1 =  dict(station_name = 'bla',
               station_id = 'blub', 
               latitude = 33, 
               longitude = 15, 
               altitude = 400,
               PI='pi1')
    d2 =  dict(station_name = 'bla',
               station_id = 'blub1', 
               latitude = 33, 
               longitude = 15, 
               altitude = 401,
               PI='pi2')
    
    assert helpers.same_meta_dict(d1, d2) == False

def test_to_pandas_timestamp():
    pass
    
def test_to_datetime64():
    pass
  
def test_is_year():
    assert helpers.is_year(2010) 
    
def test_start_stop():
    pass

def test_datetime2str():
    pass
def test_start_stop_str():
    pass

def test_start_stop_from_year():
    start, stop = helpers.start_stop_from_year(2000)
    from pandas import Timestamp
    assert start == Timestamp('2000')
    assert stop == Timestamp('2000-12-31 23:59:59')
            
def to_datestring_YYYYMMDD():
    pass

def test_cftime_to_datetime64():
    pass

def get_constraint():
    pass
    
def test_get_lat_rng_constraint():
    pass

def test_get_lon_rng_constraint():
    pass

def test_get_time_rng_constraint():
    pass   


if __name__=="__main__":
    
    test_get_standarad_name()
    test_get_standard_unit()
    test_start_stop_from_year()
    test__get_pandas_freq_and_loffset()
    test_resample_timeseries(_make_timeseries_synthetic())
    test_same_meta_dict()
    
    
    stat1 = DATA_ACCESS['station_data1']
    stat2 = DATA_ACCESS['station_data2']
    
    merged_ec = helpers.merge_station_data([stat1, stat2],
                                           var_name='ec550aer')
    
    
    
    
    
    
    
    
    